"""Pygments lexer for Scenic."""

from pygments.lexers.python import PythonLexer

class ScenicLexer(PythonLexer):
    """Lexer for Scenic code. Currently just uses the Python lexer."""
    name = 'Scenic'
    aliases = ['scenic']
    filenames = ['*.scenic', '*.sc']



"""Car models built into Webots."""

import collections

CarModel = collections.namedtuple('CarModel', ('name', 'width', 'height'))
carModels = {
	CarModel('BmwX5', 2, 4.75),
	CarModel('CitroenCZero', 1.5, 3.5),
	CarModel('LincolnMKZ', 1.9, 5),
	CarModel('RangeRoverSportSVR', 1.9, 4.5),
	CarModel('ToyotaPrius', 1.9, 4.5),
	CarModel('Bus', 3, 10),
	CarModel('Truck', 2.7, 14.5),
	CarModel('Tractor', 2, 3),
	CarModel('MotorBikeSimple', 1, 2.5)
}
modelWithName = { model.name: model for model in carModels }

smallVehicles = { modelWithName[name] for name in {
	'BmwX5', 'CitroenCZero', 'LincolnMKZ', 'RangeRoverSportSVR', 'ToyotaPrius'
}}



'''
This file contains helper functions
'''

import itertools
from typing import NamedTuple, Tuple
from copy import deepcopy

from scenic.simulators.gta.img_modf import *

'''
Compute neighbors of a given point/pixel
'''
def generate_neighbors(x):
    delta = np.arange(-1, 2)
    grid_delta = np.array(list(itertools.product(delta, delta)))
    return np.array(x)+ grid_delta


'''
Generate circle of a given radius and transformed by a center and a theta
'''
def generate_circle(radius, num_samples, theta, center):
    transformed_center = transform_center(center, radius, theta)
    thetas = np.linspace(0, 2*np.pi+0.01, num_samples)
    return transformed_center[0] + radius*np.cos(thetas), \
           transformed_center[1] + radius*np.sin(thetas)

'''
Transform the circle center as x+r*cos(90+theta) and y-r*sin(90+theta)
'''
def transform_center(center, radius, theta):
    return [center[0] + radius * np.cos(theta), center[1] - radius * np.sin(
        theta)]


'''
Compute image gradients from sobel edge detector
'''
def compute_gradient_sobel(img_data=None, img_file=None, threshold=100,
                     kernelsize=1, bw_kernelsize=3):
    assert img_data is not None or img_file is not None
    if img_data is None:
        img_data = Image.open(img_file)

    img_copy = img_data.copy()

    # Get the black and white image
    img_bw = convert_black_white(img_data=img_copy, img_file=img_file,
                                 threshold=threshold)
    cv_bw = cv2.cvtColor(np.array(img_bw), cv2.COLOR_RGB2BGR)

    # Gradients along x, y. The laplacian used for edge detection is actually
    #  a combination of these two
    sobelx = cv2.Sobel(cv_bw, cv2.CV_64F, 1, 0, ksize=kernelsize)
    sobely = cv2.Sobel(cv_bw, cv2.CV_64F, 0, 1, ksize=kernelsize)

    angle_sobel = np.arctan(sobelx / (sobely + 1e-6))

    edge_image = get_edges(img_data=img_copy, img_file=img_file,
                           threshold=threshold, kernelsize=bw_kernelsize)

    edge_x, edge_y = np.nonzero(np.asarray(edge_image))

    angle_along_edge = {}
    for x, y in zip(edge_x, edge_y):
        angle_along_edge[(x,y)] = angle_sobel[x][y][0]

    return angle_along_edge

'''
This is very coarse, avoid using this
'''
def compute_gradient_manual(img_data=None, img_file=None, threshold=100,
                     kernelsize=1):
    assert img_data is not None or img_file is not None
    if img_data is None:
        img_data = Image.open(img_file)

    img_copy = img_data.copy()

    # Directly work with the edges
    edge_image = get_edges(img_data=img_copy, img_file=img_file,
                           threshold=threshold, kernelsize=kernelsize)

    edge_x, edge_y = np.nonzero(np.asarray(edge_image))
    grad_x, grad_y = np.gradient(np.asarray(edge_image))
    angle_grad = np.arctan(grad_x/(grad_y+1e-6))

    angle_along_edge = {}
    for x, y in zip(edge_x, edge_y):
        angle_along_edge[(x,y)] = angle_grad[x][y]

    return angle_along_edge

'''
Compute all connected edges
'''

def generate_connected_edges(edge_locs):

    # This is a list of collected edges
    collected_edges = []
    while len(edge_locs) > 0:
        start_point = edge_locs.pop()
        new_edge = set({start_point})
        neighbors = set(tuple(map(tuple, generate_neighbors(start_point))))
        to_visit = neighbors.intersection(edge_locs)
        new_edge.update(neighbors.intersection(edge_locs))
        edge_locs.difference_update(new_edge.intersection(edge_locs))
        while len(to_visit) > 0:
            k = to_visit.pop()
            neighbors = set(tuple(map(tuple, generate_neighbors(k))))
            to_visit.update(neighbors.intersection(edge_locs))
            new_edge.update(neighbors.intersection(edge_locs))
            edge_locs.difference_update(new_edge.intersection(edge_locs))

        collected_edges.append(new_edge)
    return collected_edges

'''
To find the center of the road perpendicular to a given point
num_samples is function which can potentially change with the radius of the
circle
'''

def find_center(x, theta, collected_edges, all_edges,  num_samples, bw_image):

    '''
    Find which edge x lies in
    '''
    for edge in collected_edges:
        if tuple([x[1], x[0]]) in edge:
            x_edge = edge
            break

    possible_opp_edge = all_edges.difference(x_edge)

    r = 1

    '''
    Find direction of computing the growing circles
    '''
    if theta < 0:
        theta = theta + 2*np.pi
    theta_plus = theta + np.pi/2.
    theta_minus = theta - np.pi / 2.

    center_plus = transform_center(center=x, radius=2, theta=theta_plus)

    center_plus = np.array(np.around(center_plus), dtype=int)

    if center_plus[0] >= bw_image.shape[1]:
        center_plus[0] = bw_image.shape[1] - 1
    if center_plus[1] >= bw_image.shape[0]:
        center_plus[1] = bw_image.shape[0] - 1
    if bw_image[center_plus[1]][center_plus[0]] == 0:
        theta = theta_plus
        tangent_theta = theta - np.pi/2.
    else:
        theta = theta_minus
        tangent_theta = theta + np.pi / 2.

    while True:
        circle_x, circle_y = generate_circle(radius=r, center=x, theta=theta,
                                             num_samples=num_samples(r))

        int_circle = set(itertools.product(np.array(np.around(circle_y),
                        dtype=int), np.array(np.around(circle_x), dtype=int)))


        if len(possible_opp_edge.intersection(int_circle)) > 0:
            opp_edge = possible_opp_edge.intersection(int_circle)
            break

        r += 1
    if len(opp_edge) > 1:
        arg_min = np.linalg.norm(np.array(list(opp_edge)) - np.array([x[1],
                                x[0]]), axis=1).argmin()
        opp_point = np.array(list(opp_edge))[arg_min]
    else:
        opp_point = opp_edge.pop()

    mid_loc = np.array(np.around((np.array([x[1],x[0]]) +
                                  np.array(opp_point))/2.), dtype=int)

    orig_theta = tangent_theta

    # When there is a clear greater than and less than
    if mid_loc[0] > x[1] and mid_loc[1] < x[0]:
        if tangent_theta > np.pi:
            tangent_theta = tangent_theta - np.pi
        if tangent_theta < np.pi/2.:
            tangent_theta = tangent_theta + np.pi
    elif mid_loc[0] > x[1] and mid_loc[1] > x[0]:
        if tangent_theta < np.pi:
            tangent_theta = tangent_theta + np.pi
    elif mid_loc[0] < x[1] and mid_loc[1] < x[0]:
        if tangent_theta > np.pi/2.:
            tangent_theta = tangent_theta - np.pi
    elif mid_loc[0] < x[1] and mid_loc[1] > x[0]:
        if tangent_theta < 3.*np.pi/2. and tangent_theta > 0:
            tangent_theta = tangent_theta + np.pi


    # When there exists one equality
    if mid_loc[0] == x[1]:
        if mid_loc[1] < x[0]:
            if tangent_theta > np.pi:
                tangent_theta = tangent_theta - np.pi
        else:
            if tangent_theta < 3./2.*np.pi:
                tangent_theta = tangent_theta + np.pi

    if mid_loc[1] == x[0]:
        if mid_loc[0] < x[1]:
            if tangent_theta > np.pi/2.:
                tangent_theta = tangent_theta - np.pi
        else:
            if tangent_theta < np.pi:
                tangent_theta = tangent_theta + np.pi


    if tangent_theta < 0:
        tangent_theta = tangent_theta + 2 * np.pi




    return orig_theta, tangent_theta, tuple(opp_point), tuple(mid_loc)


'''
Update all edges with the bounding box of the image
'''

def compute_bb(img_dimensions):
    im_x_min = -1
    im_y_min = -1
    im_x_max = img_dimensions[1]
    im_y_max = img_dimensions[0]

    '''
    Compute the bounding box
    '''
    bb = set()
    bb.update({(im_x_min, y) for y in range(im_y_min, im_y_max+1)})
    bb.update({(im_x_max, y) for y in range(im_y_min, im_y_max + 1)})
    bb.update({(x, im_y_min) for x in range(im_x_min, im_x_max + 1)})
    bb.update({(x, im_y_max) for x in range(im_x_min, im_x_max + 1)})

    return bb

class EdgeData(NamedTuple):
    init_theta: float
    tangent: float
    opp_loc: Tuple[float, float]
    mid_loc: Tuple[float, float]

'''
Compute the dictionary with key is the start state and the values are a tuple
consisting of the angle, mid point and the the opposite edge
'''

def compute_midpoints(img_data, bw_kernelsize=1, kernelsize=3, threshold=100,
                      num_samples = lambda r:16):
    # First compute the black and white image and store it
    img_bw = convert_black_white(img_data=img_data)
    img_bw = img_bw.convert('L')
    img_bw_int = np.array(img_bw, dtype=int)

    # Compute the point to the tangent angle dictionary
    cae = compute_gradient_sobel(img_data=img_data,
                                 bw_kernelsize=bw_kernelsize,
                                 kernelsize=kernelsize,
                                 threshold=threshold)

    # Collect all edges
    all_edges = set(cae.keys())

    # Collect all roads
    connected_edges = generate_connected_edges(edge_locs=deepcopy(all_edges))

    # Compute bounding box
    bb = compute_bb(img_dimensions=img_data.size)

    all_edges.update(bb)

    # Compute midpoints
    point_data = {}
    for edge in connected_edges:
        for x in edge:
            init_theta, theta, opp_loc, mid_loc = find_center(x=[x[1], x[0]],
                                                    theta=cae[x],
                                           collected_edges=connected_edges,
                                           all_edges=all_edges,
                                           num_samples=num_samples,
                                           bw_image=img_bw_int)

            point_data[x] = EdgeData(init_theta=init_theta, tangent=theta,
                             opp_loc=opp_loc, mid_loc=mid_loc \
                    if opp_loc not in bb else (np.nan, np.nan))

    return point_data

'''
Compute the heading at a random sample
'''

def compute_heading(x, point_data):
    road_edges = tuple(point_data.keys())
    edge_distance = np.linalg.norm(np.array(list(road_edges)) - np.array(x),1,
                                   axis=1)
    closest_edge = list(road_edges)[edge_distance.argmin()]

    return closest_edge, point_data[closest_edge]

'''
Sampling from road
'''

def sample_from_road(img_data, num_samples):
    img_bw = convert_black_white(img_data=img_data)
    img_bw = img_bw.convert('L')
    img_bw_int = np.array(img_bw, dtype=int)

    road_y, road_x = np.where(img_bw_int == 0)
    all_roads = list(zip(road_y, road_x))

    locs = np.random.choice(len(all_roads), num_samples, replace=False)

    return np.array(all_roads)[locs]



"""Support for checking Scenic types."""

import sys
import inspect

import numpy as np

from scenic.core.distributions import Distribution
from scenic.core.lazy_eval import (DelayedArgument, valueInContext, requiredProperties,
                                   needsLazyEvaluation, toDelayedArgument)
from scenic.core.vectors import Vector
from scenic.core.utils import RuntimeParseError

# Typing and coercion rules:
#
# coercible to a scalar:
#   float
#   int (by conversion to float)
#	numpy real scalar types (likewise)
# coercible to a heading:
#	anything coercible to a scalar
#	anything with a toHeading() method
# coercible to a Vector:
#   anything with a toVector() method
# coercible to an object of type T:
#   instances of T
#
# Finally, Distributions are coercible to T iff their valueType is.

## Basic types

class Heading:
	"""Dummy class used as a target for type coercions to headings."""
	pass

def underlyingType(thing):
	"""What type this value ultimately evaluates to, if we can tell."""
	if isinstance(thing, Distribution):
		return thing.valueType
	elif isinstance(thing, TypeChecker) and len(thing.types) == 1:
		return thing.types[0]
	else:
		return type(thing)

def isA(thing, ty):
	"""Does this evaluate to a member of the given Scenic type?"""
	return issubclass(underlyingType(thing), ty)

def unifyingType(opts):		# TODO improve?
	"""Most specific type unifying the given types."""
	types = [underlyingType(opt) for opt in opts]
	if all(issubclass(ty, (float, int)) for ty in types):
		return float
	mro = inspect.getmro(types[0])
	for parent in mro:
		if all(issubclass(ty, parent) for ty in types):
			return parent
	raise RuntimeError(f'broken MRO for types {types}')

## Type coercions (for internal use -- see the type checking API below)

def canCoerceType(typeA, typeB):
	"""Can values of typeA be coerced into typeB?"""
	if typeB is float:
		if issubclass(typeA, (float, int)):
			return True
		if issubclass(typeA, np.number) and not issubclass(typeA, np.complexfloating):
			return True
		return False
	elif typeB is Heading:
		return canCoerceType(typeA, float) or hasattr(typeA, 'toHeading')
	elif typeB is Vector:
		return hasattr(typeA, 'toVector')
	else:
		return issubclass(typeA, typeB)

def canCoerce(thing, ty):
	"""Can this value be coerced into the given type?"""
	tt = underlyingType(thing)
	return canCoerceType(tt, ty)

def coerce(thing, ty):
	"""Coerce something into the given type."""
	assert canCoerce(thing, ty), (thing, ty)
	if isinstance(thing, Distribution):
		return thing
	if ty is float:
		return float(thing)
	elif ty is Heading:
		return thing.toHeading() if hasattr(thing, 'toHeading') else float(thing)
	elif ty is Vector:
		return thing.toVector()
	else:
		return thing

def coerceToAny(thing, types, error):
	"""Coerce something into any of the given types, printing an error if impossible."""
	for ty in types:
		if canCoerce(thing, ty):
			return coerce(thing, ty)
	print(f'Failed to coerce {thing} of type {underlyingType(thing)} to {types}', file=sys.stderr)
	raise RuntimeParseError(error)

## Top-level type checking/conversion API

def toTypes(thing, types, typeError='wrong type'):
	"""Convert something to any of the given types, printing an error if impossible."""
	if needsLazyEvaluation(thing):
		# cannot check the type now; create proxy object to check type after evaluation
		return TypeChecker(thing, types, typeError)
	else:
		return coerceToAny(thing, types, typeError)

def toType(thing, ty, typeError='wrong type'):
	"""Convert something to a given type, printing an error if impossible."""
	return toTypes(thing, (ty,), typeError)

def toScalar(thing, typeError='non-scalar in scalar context'):
	"""Convert something to a scalar, printing an error if impossible."""
	return toType(thing, float, typeError)

def toHeading(thing, typeError='non-heading in heading context'):
	"""Convert something to a heading, printing an error if impossible."""
	return toType(thing, Heading, typeError)

def toVector(thing, typeError='non-vector in vector context'):
	"""Convert something to a vector, printing an error if impossible."""
	return toType(thing, Vector, typeError)

def evaluateRequiringEqualTypes(func, thingA, thingB, typeError='type mismatch'):
	"""Evaluate the func, assuming thingA and thingB have the same type.

	If func produces a lazy value, it should not have any required properties beyond
	those of thingA and thingB."""
	if not needsLazyEvaluation(thingA) and not needsLazyEvaluation(thingB):
		if underlyingType(thingA) is not underlyingType(thingB):
			raise RuntimeParseError(typeError)
		return func()
	else:
		# cannot check the types now; create proxy object to check types after evaluation
		return TypeEqualityChecker(func, thingA, thingB, typeError)

## Proxy objects for lazy type checking

class TypeChecker(DelayedArgument):
	"""Checks that a given lazy value has one of a given list of types."""
	def __init__(self, arg, types, error):
		def check(context):
			val = arg.evaluateIn(context)
			return coerceToAny(val, types, error)
		super().__init__(requiredProperties(arg), check)
		self.inner = arg
		self.types = types

	def __str__(self):
		return f'TypeChecker({self.inner},{self.types})'

class TypeEqualityChecker(DelayedArgument):
	"""Lazily evaluates a function, after checking that two lazy values have the same type."""
	def __init__(self, func, checkA, checkB, error):
		props = requiredProperties(checkA) | requiredProperties(checkB)
		def check(context):
			ca = valueInContext(checkA, context)
			cb = valueInContext(checkB, context)
			if underlyingType(ca) is not underlyingType(cb):
				raise RuntimeParseError(error)
			return valueInContext(func(), context)
		super().__init__(props, check)
		self.inner = func
		self.checkA = checkA
		self.checkB = checkB

	def __str__(self):
		return f'TypeEqualityChecker({self.inner},{self.checkA},{self.checkB})'




"""Support for lazy evaluation of expressions and specifiers."""

import itertools

class LazilyEvaluable:
	"""Values which may require evaluation in the context of an object being constructed.

	If a LazilyEvaluable specifies any properties it depends on, then it cannot be evaluated to a
	normal value except during the construction of an object which already has values for those
	properties.
	"""
	def __init__(self, requiredProps):
		self._dependencies = ()		# TODO improve?
		self._requiredProperties = set(requiredProps)

	def evaluateIn(self, context):
		"""Evaluate this value in the context of an object being constructed.

		The object must define all of the properties on which this value depends.
		"""
		assert all(hasattr(context, prop) for prop in self._requiredProperties)
		value = self.evaluateInner(context)
		assert not needsLazyEvaluation(value)	# value should not require further evaluation
		return value

	def evaluateInner(self, context):
		"""Actually evaluate in the given context, which provides all required properties."""
		return self

class DelayedArgument(LazilyEvaluable):
	"""Specifier arguments requiring other properties to be evaluated first.

	The value of a DelayedArgument is given by a function mapping the context (object under
	construction) to a value.
	"""
	def __init__(self, requiredProps, value):
		self.value = value
		super().__init__(requiredProps)

	def evaluateInner(self, context):
		return self.value(context)

	def __getattr__(self, name):
		return DelayedArgument(self._requiredProperties,
			lambda context: getattr(self.evaluateIn(context), name))

	def __call__(self, *args, **kwargs):
		dargs = [toDelayedArgument(arg) for arg in args]
		kwdargs = { name: toDelayedArgument(arg) for name, arg in kwargs.items() }
		subprops = (darg._requiredProperties for darg in itertools.chain(dargs, kwdargs.values()))
		props = self._requiredProperties.union(*subprops)
		def value(context):
			subvalues = (darg.evaluateIn(context) for darg in dargs)
			kwsvs = { name: darg.evaluateIn(context) for name, darg in kwdargs.items() }
			return self.evaluateIn(context)(*subvalues, **kwsvs)
		return DelayedArgument(props, value)

# Operators which can be applied to DelayedArguments
allowedOperators = [
	'__neg__',
	'__pos__',
	'__abs__',
	'__lt__', '__le__',
	'__eq__', '__ne__',
	'__gt__', '__ge__',
	'__add__', '__radd__',
	'__sub__', '__rsub__',
	'__mul__', '__rmul__',
	'__truediv__', '__rtruediv__',
	'__floordiv__', '__rfloordiv__',
	'__mod__', '__rmod__',
	'__divmod__', '__rdivmod__',
	'__pow__', '__rpow__',
	'__round__',
	'__len__',
	'__getitem__'
]
def makeDelayedOperatorHandler(op):
	def handler(self, *args):
		dargs = [toDelayedArgument(arg) for arg in args]
		props = self._requiredProperties.union(*(darg._requiredProperties for darg in dargs))
		def value(context):
			subvalues = (darg.evaluateIn(context) for darg in dargs)
			return getattr(self.evaluateIn(context), op)(*subvalues)
		return DelayedArgument(props, value)
	return handler
for op in allowedOperators:
	setattr(DelayedArgument, op, makeDelayedOperatorHandler(op))

def makeDelayedFunctionCall(func, args, kwargs):
	"""Utility function for creating a lazily-evaluated function call."""
	dargs = [toDelayedArgument(arg) for arg in args]
	kwdargs = { name: toDelayedArgument(arg) for name, arg in kwargs.items() }
	props = set().union(*(darg._requiredProperties
	                      for darg in itertools.chain(dargs, kwdargs.values())))
	def value(context):
		subvalues = (darg.evaluateIn(context) for darg in dargs)
		kwsubvals = { name: darg.evaluateIn(context) for name, darg in kwdargs.items() }
		return func(*subvalues, **kwsubvals)
	return DelayedArgument(props, value)

def valueInContext(value, context):
	"""Evaluate something in the context of an object being constructed."""
	try:
		return value.evaluateIn(context)
	except AttributeError:
		return value

def toDelayedArgument(thing):
	if isinstance(thing, DelayedArgument):
		return thing
	return DelayedArgument(set(), lambda context: thing)

def requiredProperties(thing):
	if hasattr(thing, '_requiredProperties'):
		return thing._requiredProperties
	return set()

def needsLazyEvaluation(thing):
	return isinstance(thing, DelayedArgument) or requiredProperties(thing)



"""Implementations of the built-in Scenic classes."""

import inspect
import collections
import math
import random

from scenic.core.distributions import Samplable, needsSampling
from scenic.core.specifiers import Specifier, PropertyDefault
from scenic.core.vectors import Vector
from scenic.core.geometry import RotatedRectangle, averageVectors, hypot, min, pointIsInCone
from scenic.core.regions import CircularRegion, SectorRegion
from scenic.core.type_support import toVector, toScalar
from scenic.core.lazy_eval import needsLazyEvaluation
from scenic.core.utils import areEquivalent, RuntimeParseError

## Abstract base class

class Constructible(Samplable):
	"""Abstract base class for Scenic objects.

	Scenic objects, which are constructed using specifiers, are implemented
	internally as instances of ordinary Python classes. This abstract class
	implements the procedure to resolve specifiers and determine values for
	the properties of an object, as well as several common methods supported
	by objects.
	"""

	@classmethod
	def defaults(cla):		# TODO improve so this need only be done once?
		# find all defaults provided by the class or its superclasses
		allDefs = collections.defaultdict(list)
		for sc in inspect.getmro(cla):
			if hasattr(sc, '__annotations__'):
				for prop, value in sc.__annotations__.items():
					allDefs[prop].append(PropertyDefault.forValue(value))

		# resolve conflicting defaults
		resolvedDefs = {}
		for prop, defs in allDefs.items():
			primary, rest = defs[0], defs[1:]
			spec = primary.resolveFor(prop, rest)
			resolvedDefs[prop] = spec
		return resolvedDefs

	@classmethod
	def withProperties(cls, props):
		assert all(reqProp in props for reqProp in cls.defaults())
		assert all(not needsLazyEvaluation(val) for val in props.values())
		specs = (Specifier(prop, val) for prop, val in props.items())
		return cls(*specs)

	def __init__(self, *args, **kwargs):
		# Validate specifiers
		name = type(self).__name__
		specifiers = list(args)
		for prop, val in kwargs.items():
			specifiers.append(Specifier(prop, val))
		properties = dict()
		optionals = collections.defaultdict(list)
		defs = self.defaults()
		for spec in specifiers:
			assert isinstance(spec, Specifier), (name, spec)
			prop = spec.property
			if prop in properties:
				raise RuntimeParseError(f'property "{prop}" of {name} specified twice')
			properties[prop] = spec
			for opt in spec.optionals:
				if opt in defs:		# do not apply optionals for properties this object lacks
					optionals[opt].append(spec)

		# Decide which optionals to use
		optionalsForSpec = collections.defaultdict(set)
		for opt, specs in optionals.items():
			if opt in properties:
				continue		# optionals do not override a primary specification
			if len(specs) > 1:
				raise RuntimeParseError(f'property "{opt}" of {name} specified twice (optionally)')
			assert len(specs) == 1
			spec = specs[0]
			properties[opt] = spec
			optionalsForSpec[spec].add(opt)

		# Add any default specifiers needed
		for prop in defs:
			if prop not in properties:
				spec = defs[prop]
				specifiers.append(spec)
				properties[prop] = spec

		# Topologically sort specifiers
		order = []
		seen, done = set(), set()

		def dfs(spec):
			if spec in done:
				return
			elif spec in seen:
				raise RuntimeParseError(f'specifier for property {spec.property} '
										'depends on itself')
			seen.add(spec)
			for dep in spec.requiredProperties:
				child = properties.get(dep)
				if child is None:
					raise RuntimeParseError(f'property {dep} required by '
											f'specifier {spec} is not specified')
				else:
					dfs(child)
			order.append(spec)
			done.add(spec)

		for spec in specifiers:
			dfs(spec)
		assert len(order) == len(specifiers)

		# Evaluate and apply specifiers
		for spec in order:
			spec.applyTo(self, optionalsForSpec[spec])

		# Set up dependencies
		deps = []
		for prop in properties:
			assert hasattr(self, prop)
			val = getattr(self, prop)
			if needsSampling(val):
				deps.append(val)
		super().__init__(deps)
		self.properties = set(properties)

	def sampleGiven(self, value):
		return self.withProperties({ prop: value[getattr(self, prop)]
								   for prop in self.properties })

	def allProperties(self):
		return { prop: getattr(self, prop) for prop in self.properties }

	def copyWith(self, **overrides):
		props = self.allProperties()
		props.update(overrides)
		return self.withProperties(props)

	def isEquivalentTo(self, other):
		if type(other) is not type(self):
			return False
		return areEquivalent(self.allProperties(), other.allProperties())

	def __str__(self):
		if hasattr(self, 'properties'):
			allProps = { prop: getattr(self, prop) for prop in self.properties }
		else:
			allProps = '<under construction>'
		return f'{type(self).__name__}({allProps})'

## Mutators

class Mutator:
	"""An object controlling how the ``mutate`` statement affects an `Object`.

	A `Mutator` can be assigned to the ``mutator`` property of an `Object` to
	control the effect of the ``mutate`` statement. When mutation is enabled
	for such an object using that statement, the mutator's `appliedTo` method
	is called to compute a mutated version.
	"""

	def appliedTo(self, obj):
		"""Return a mutated copy of the object. Implemented by subclasses."""
		raise NotImplementedError

class PositionMutator(Mutator):
	"""Mutator adding Gaussian noise to ``position``. Used by `Point`.

	Attributes:
		stddev (float): standard deviation of noise
	"""
	def __init__(self, stddev):
		self.stddev = stddev

	def appliedTo(self, obj):
		noise = Vector(random.gauss(0, self.stddev), random.gauss(0, self.stddev))
		pos = toVector(obj.position, '"position" not a vector')
		pos = pos + noise
		return (obj.copyWith(position=pos), True)		# allow further mutation

	def __eq__(self, other):
		if type(other) is not type(self):
			return NotImplemented
		return (other.stddev == self.stddev)

	def __hash__(self):
		return hash(self.stddev)

class HeadingMutator(Mutator):
	"""Mutator adding Gaussian noise to ``heading``. Used by `OrientedPoint`.

	Attributes:
		stddev (float): standard deviation of noise
	"""
	def __init__(self, stddev):
		self.stddev = stddev

	def appliedTo(self, obj):
		noise = random.gauss(0, self.stddev)
		h = obj.heading + noise
		return (obj.copyWith(heading=h), True)		# allow further mutation

	def __eq__(self, other):
		if type(other) is not type(self):
			return NotImplemented
		return (other.stddev == self.stddev)

	def __hash__(self):
		return hash(self.stddev)

## Point

class Point(Constructible):
	"""Implementation of the Scenic class ``Point``.

	The default mutator for `Point` adds Gaussian noise to ``position`` with
	a standard deviation given by the ``positionStdDev`` property.

	Attributes:
		position (`Vector`): Position of the point. Default value is the origin.
		visibleDistance (float): Distance for ``can see`` operator. Default value 50.
		width (float): Default value zero (only provided for compatibility with
		  operators that expect an `Object`).
		height (float): Default value zero.
	"""
	position: Vector(0, 0)
	width: 0
	height: 0
	visibleDistance: 50

	mutationEnabled: False
	mutator: PropertyDefault({'positionStdDev'}, {'additive'},
							 lambda self: PositionMutator(self.positionStdDev))
	positionStdDev: 1

	def __init__(self, *args, **kwargs):
		super().__init__(*args, **kwargs)
		self.position = toVector(self.position, f'"position" of {self} not a vector')
		self.corners = (self.position,)
		self.visibleRegion = CircularRegion(self.position, self.visibleDistance)

	def toVector(self):
		return self.position.toVector()

	def canSee(self, other):	# TODO improve approximation?
		for corner in other.corners:
			if self.visibleRegion.containsPoint(corner):
				return True
		return False

	def sampleGiven(self, value):
		sample = super().sampleGiven(value)
		if self.mutationEnabled:
			for mutator in self.mutator:
				if mutator is None:
					continue
				sample, proceed = mutator.appliedTo(sample)
				if not proceed:
					break
		return sample

	# Points automatically convert to Vectors when needed
	def __getattr__(self, attr):
		if hasattr(Vector, attr):
			return getattr(self.toVector(), attr)
		else:
			raise AttributeError(f"'{type(self).__name__}' object has no attribute '{attr}'")

## OrientedPoint

class OrientedPoint(Point):
	"""Implementation of the Scenic class ``OrientedPoint``.

	The default mutator for `OrientedPoint` adds Gaussian noise to ``heading``
	with a standard deviation given by the ``headingStdDev`` property, then
	applies the mutator for `Point`.

	Attributes:
		heading (float): Heading of the `OrientedPoint`. Default value 0 (North).
		viewAngle (float): View cone angle for ``can see`` operator. Default
		  value :math:`2\\pi`.
	"""
	heading: 0
	viewAngle: math.tau

	mutator: PropertyDefault({'headingStdDev'}, {'additive'},
		lambda self: HeadingMutator(self.headingStdDev))
	headingStdDev: math.radians(5)

	def __init__(self, *args, **kwargs):
		super().__init__(*args, **kwargs)
		self.heading = toScalar(self.heading, f'"heading" of {self} not a scalar')
		self.visibleRegion = SectorRegion(self.position, self.visibleDistance,
										  self.heading, self.viewAngle)

	def relativize(self, vec):
		pos = self.relativePosition(vec)
		return OrientedPoint(position=pos, heading=self.heading)

	def relativePosition(self, x, y=None):
		vec = x if y is None else Vector(x, y)
		pos = self.position.offsetRotated(self.heading, vec)
		return OrientedPoint(position=pos, heading=self.heading)

	def toHeading(self):
		return self.heading

## Object

class Object(OrientedPoint, RotatedRectangle):
	"""Implementation of the Scenic class ``Object``.

	Attributes:
		width (float): Width of the object, i.e. extent along its X axis.
		  Default value 1.
		height (float): Height of the object, i.e. extent along its Y axis.
		  Default value 1.
		allowCollisions (bool): Whether the object is allowed to intersect
		  other objects. Default value ``False``.
		requireVisible (bool): Whether the object is required to be visible
		  from the ``ego`` object. Default value ``True``.
		regionContainedIn (`Region` or ``None``): A `Region` the object is
		  required to be contained in. If ``None``, the object need only be
		  contained in the scenario's workspace.
		cameraOffset (`Vector`): Position of the camera for the ``can see``
		  operator, relative to the object's ``position``. Default ``0 @ 0``.
	"""
	width: 1
	height: 1
	allowCollisions: False
	requireVisible: True
	regionContainedIn: None
	cameraOffset: Vector(0, 0)

	def __init__(self, *args, **kwargs):
		super().__init__(*args, **kwargs)
		import scenic.syntax.veneer as veneer	# TODO improve?
		veneer.registerObject(self)
		self.hw = hw = self.width / 2
		self.hh = hh = self.height / 2
		self.radius = hypot(hw, hh)	# circumcircle; for collision detection
		self.inradius = min(hw, hh)	# incircle; for collision detection
		self.left = self.relativePosition(-hw, 0)
		self.right = self.relativePosition(hw, 0)
		self.front = self.relativePosition(0, hh)
		self.back = self.relativePosition(0, -hh)
		self.frontLeft = self.relativePosition(-hw, hh)
		self.frontRight = self.relativePosition(hw, hh)
		self.backLeft = self.relativePosition(-hw, -hh)
		self.backRight = self.relativePosition(hw, -hh)
		self.corners = (self.frontRight.toVector(), self.frontLeft.toVector(),
			self.backLeft.toVector(), self.backRight.toVector())
		camera = self.position.offsetRotated(self.heading, self.cameraOffset)
		self.visibleRegion = SectorRegion(camera, self.visibleDistance,
										  self.heading, self.viewAngle)
		self._relations = []

	def show(self, workspace, plt, highlight=False):
		if needsSampling(self):
			raise RuntimeError('tried to show() symbolic Object')
		pos = self.position
		spos = workspace.scenicToSchematicCoords(pos)

		if highlight:
			# Circle around object
			rad = 1.5 * max(self.width, self.height)
			c = plt.Circle(spos, rad, color='g', fill=False)
			plt.gca().add_artist(c)
			# View cone
			ha = self.viewAngle / 2.0
			camera = self.position.offsetRotated(self.heading, self.cameraOffset)
			cpos = workspace.scenicToSchematicCoords(camera)
			for angle in (-ha, ha):
				p = camera.offsetRadially(20, self.heading + angle)
				edge = [cpos, workspace.scenicToSchematicCoords(p)]
				x, y = zip(*edge)
				plt.plot(x, y, 'b:')

		corners = [workspace.scenicToSchematicCoords(corner) for corner in self.corners]
		x, y = zip(*corners)
		color = self.color if hasattr(self, 'color') else (1, 0, 0)
		plt.fill(x, y, color=color)

		frontMid = averageVectors(corners[0], corners[1])
		baseTriangle = [frontMid, corners[2], corners[3]]
		triangle = [averageVectors(p, spos, weight=0.5) for p in baseTriangle]
		x, y = zip(*triangle)
		plt.fill(x, y, "w")
		plt.plot(x + (x[0],), y + (y[0],), color="k", linewidth=1)



"""Stub to allow changing the map without having to change the model."""

import os.path

mapPath = None
lanePoints = 20

def setMapPath(module, relpath, points=None):
    global mapPath, lanePoints
    base = os.path.dirname(module)
    mapPath = os.path.join(base, relpath)
    if points is not None:
        lanePoints = points




# stub to allow changing the intersection without having to alter guideways_model.sc

import os.path

intersectionPath = None
intersectionOrigin = None

def setLocalIntersection(module, relpath):
    global intersectionPath
    base = os.path.dirname(module)
    intersectionPath = os.path.join(base, relpath)



"""World model and associated code for traffic scenarios in Webots.

This model handles Webots world files generated from Open Street Map data using the Webots OSM importer.

.. raw:: html

   <h2>Submodules</h2>

.. autosummary::
   :toctree: _autosummary

   model
   world
   interface
   car_models
"""




"""Scenic vectors and vector fields."""

import math
from math import sin, cos
import random
import collections
import itertools
import functools

import shapely.geometry

from scenic.core.distributions import (Samplable, Distribution, MethodDistribution,
                                       needsSampling, makeOperatorHandler, distributionMethod)
from scenic.core.lazy_eval import valueInContext, needsLazyEvaluation, makeDelayedFunctionCall
import scenic.core.utils as utils
from scenic.core.geometry import normalizeAngle

class VectorDistribution(Distribution):
	"""A distribution over Vectors."""
	defaultValueType = None		# will be set after Vector is defined

	def toVector(self):
		return self

class CustomVectorDistribution(VectorDistribution):
	"""Distribution with a custom sampler given by an arbitrary function."""
	def __init__(self, sampler, *dependencies, name='CustomVectorDistribution', evaluator=None):
		super().__init__(*dependencies)
		self.sampler = sampler
		self.name = name
		self.evaluator = evaluator

	def sampleGiven(self, value):
		return self.sampler(value)

	def evaluateInner(self, context):
		if self.evaluator is None:
			raise NotImplementedError('evaluateIn() not supported by this distribution')
		return self.evaluator(self, context)

	def __str__(self):
		deps = utils.argsToString(self.dependencies)
		return f'{self.name}{deps}'

class VectorOperatorDistribution(VectorDistribution):
	"""Vector version of OperatorDistribution."""
	def __init__(self, operator, obj, operands):
		super().__init__(obj, *operands)
		self.operator = operator
		self.object = obj
		self.operands = operands

	def sampleGiven(self, value):
		first = value[self.object]
		rest = (value[child] for child in self.operands)
		op = getattr(first, self.operator)
		return op(*rest)

	def evaluateInner(self, context):
		obj = valueInContext(self.object, context)
		operands = tuple(valueInContext(arg, context) for arg in self.operands)
		return VectorOperatorDistribution(self.operator, obj, operands)

	def __str__(self):
		ops = utils.argsToString(self.operands)
		return f'{self.object}.{self.operator}{ops}'

class VectorMethodDistribution(VectorDistribution):
	"""Vector version of MethodDistribution."""
	def __init__(self, method, obj, args, kwargs):
		super().__init__(*args, *kwargs.values())
		self.method = method
		self.object = obj
		self.arguments = args
		self.kwargs = kwargs

	def sampleGiven(self, value):
		args = (value[arg] for arg in self.arguments)
		kwargs = { name: value[arg] for name, arg in self.kwargs.items() }
		return self.method(self.object, *args, **kwargs)

	def evaluateInner(self, context):
		obj = valueInContext(self.object, context)
		arguments = tuple(valueInContext(arg, context) for arg in self.arguments)
		kwargs = { name: valueInContext(arg, context) for name, arg in self.kwargs.items() }
		return VectorMethodDistribution(self.method, obj, arguments, kwargs)

	def __str__(self):
		args = utils.argsToString(itertools.chain(self.arguments, self.kwargs.values()))
		return f'{self.object}.{self.method.__name__}{args}'

def scalarOperator(method):
	"""Decorator for vector operators that yield scalars."""
	op = method.__name__
	setattr(VectorDistribution, op, makeOperatorHandler(op))

	@functools.wraps(method)
	def handler2(self, *args, **kwargs):
		if any(needsSampling(arg) for arg in itertools.chain(args, kwargs.values())):
			return MethodDistribution(method, self, args, kwargs)
		else:
			return method(self, *args, **kwargs)
	return handler2

def makeVectorOperatorHandler(op):
	def handler(self, *args):
		return VectorOperatorDistribution(op, self, args)
	return handler
def vectorOperator(method):
	"""Decorator for vector operators that yield vectors."""
	op = method.__name__
	setattr(VectorDistribution, op, makeVectorOperatorHandler(op))

	@functools.wraps(method)
	def handler2(self, *args):
		if needsSampling(self):
			return VectorOperatorDistribution(op, self, args)
		elif any(needsSampling(arg) for arg in args):
			return VectorMethodDistribution(method, self, args, {})
		elif any(needsLazyEvaluation(arg) for arg in args):
			# see analogous comment in distributionFunction
			return makeDelayedFunctionCall(handler2, args, {})
		else:
			return method(self, *args)
	return handler2

def vectorDistributionMethod(method):
	"""Decorator for methods that produce vectors. See distributionMethod."""
	@functools.wraps(method)
	def helper(self, *args, **kwargs):
		if any(needsSampling(arg) for arg in itertools.chain(args, kwargs.values())):
			return VectorMethodDistribution(method, self, args, kwargs)
		elif any(needsLazyEvaluation(arg) for arg in itertools.chain(args, kwargs.values())):
			# see analogous comment in distributionFunction
			return makeDelayedFunctionCall(helper, (self,) + args, kwargs)
		else:
			return method(self, *args, **kwargs)
	return helper

class Vector(Samplable, collections.abc.Sequence):
	"""A 2D vector, whose coordinates can be distributions."""
	def __init__(self, x, y):
		self.coordinates = (x, y)
		super().__init__(self.coordinates)

	@property
	def x(self):
		return self.coordinates[0]

	@property
	def y(self):
		return self.coordinates[1]

	def toVector(self):
		return self

	def sampleGiven(self, value):
		return Vector(*(value[coord] for coord in self.coordinates))

	def evaluateInner(self, context):
		return Vector(*(valueInContext(coord, context) for coord in self.coordinates))

	@vectorOperator
	def rotatedBy(self, angle):
		"""Return a vector equal to this one rotated counterclockwise by the given angle."""
		x, y = self.x, self.y
		c, s = cos(angle), sin(angle)
		return Vector((c * x) - (s * y), (s * x) + (c * y))

	@vectorOperator
	def offsetRotated(self, heading, offset):
		ro = offset.rotatedBy(heading)
		return self + ro

	@vectorOperator
	def offsetRadially(self, radius, heading):
		return self.offsetRotated(heading, Vector(0, radius))

	@scalarOperator
	def distanceTo(self, other):
		dx, dy = other.toVector() - self
		return math.hypot(dx, dy)

	@scalarOperator
	def angleTo(self, other):
		dx, dy = other.toVector() - self
		return normalizeAngle(math.atan2(dy, dx) - (math.pi / 2))

	@vectorOperator
	def __add__(self, other):
		return Vector(self[0] + other[0], self[1] + other[1])

	@vectorOperator
	def __radd__(self, other):
		return Vector(self[0] + other[0], self[1] + other[1])

	@vectorOperator
	def __sub__(self, other):
		return Vector(self[0] - other[0], self[1] - other[1])

	@vectorOperator
	def __rsub__(self, other):
		return Vector(other[0] - self[0], other[1] - self[1])

	def __len__(self):
		return len(self.coordinates)

	def __getitem__(self, index):
		return self.coordinates[index]

	def __repr__(self):
		return f'({self.x} @ {self.y})'

	def __eq__(self, other):
		if type(other) is not Vector:
			return NotImplemented
		return other.coordinates == self.coordinates

	def __hash__(self):
		return hash(self.coordinates)

VectorDistribution.defaultValueType = Vector

class OrientedVector(Vector):
	def __init__(self, x, y, heading):
		super().__init__(x, y)
		self.heading = heading

	def toHeading(self):
		return self.heading

	def __eq__(self, other):
		if type(other) is not OrientedVector:
			return NotImplemented
		return (other.coordinates == self.coordinates
		    and other.heading == self.heading)

	def __hash__(self):
		return hash((self.coordinates, self.heading))

class VectorField:
	def __init__(self, name, value):
		self.name = name
		self.value = value
		self.valueType = float

	@distributionMethod
	def __getitem__(self, pos):
		return self.value(pos)

	@vectorDistributionMethod
	def followFrom(self, pos, dist, steps=4):
		step = dist / steps
		for i in range(steps):
			pos = pos.offsetRadially(step, self[pos])
		return pos

	def __str__(self):
		return f'<{type(self).__name__} {self.name}>'

class PolygonalVectorField(VectorField):
	def __init__(self, name, cells, headingFunction=None, defaultHeading=None):
		self.cells = tuple(cells)
		if headingFunction is None and defaultHeading is not None:
			headingFunction = lambda pos: defaultHeading
		self.headingFunction = headingFunction
		for cell, heading in self.cells:
			if heading is None and headingFunction is None and defaultHeading is None:
				raise RuntimeError(f'missing heading for cell of PolygonalVectorField')
		self.defaultHeading = defaultHeading
		super().__init__(name, self.valueAt)

	def valueAt(self, pos):
		point = shapely.geometry.Point(pos)
		for cell, heading in self.cells:
			if cell.intersects(point):
				return self.headingFunction(pos) if heading is None else heading
		if self.defaultHeading is not None:
			return self.defaultHeading
		raise RuntimeError(f'evaluated PolygonalVectorField at undefined point {pos}')



"""Objects representing regions in space."""

import math
import random
import itertools

import numpy
import scipy.spatial
import shapely.geometry
import shapely.ops

from scenic.core.distributions import Samplable, RejectionException, needsSampling
from scenic.core.lazy_eval import valueInContext
from scenic.core.vectors import Vector, OrientedVector, VectorDistribution
from scenic.core.geometry import RotatedRectangle
from scenic.core.geometry import sin, cos, hypot, findMinMax, pointIsInCone, averageVectors
from scenic.core.geometry import headingOfSegment, triangulatePolygon, plotPolygon, polygonUnion
from scenic.core.type_support import toVector
from scenic.core.utils import cached, areEquivalent

def toPolygon(thing):
	if needsSampling(thing):
		return None
	if hasattr(thing, 'polygon'):
		return thing.polygon
	if hasattr(thing, 'polygons'):
		return thing.polygons
	if hasattr(thing, 'lineString'):
		return thing.lineString
	return None

def regionFromShapelyObject(obj, orientation=None):
	"""Build a 'Region' from Shapely geometry."""
	assert obj.is_valid, obj
	if obj.is_empty:
		return nowhere
	elif isinstance(obj, (shapely.geometry.Polygon, shapely.geometry.MultiPolygon)):
		return PolygonalRegion(polygon=obj, orientation=orientation)
	elif isinstance(obj, (shapely.geometry.LineString, shapely.geometry.MultiLineString)):
		return PolylineRegion(polyline=obj, orientation=orientation)
	else:
		raise RuntimeError(f'unhandled type of Shapely geometry: {obj}')

class PointInRegionDistribution(VectorDistribution):
	"""Uniform distribution over points in a Region"""
	def __init__(self, region):
		super().__init__(region)
		self.region = region

	def sampleGiven(self, value):
		return value[self.region].uniformPointInner()

	def __str__(self):
		return f'PointIn({self.region})'

class Region(Samplable):
	"""Abstract class for regions."""
	def __init__(self, name, *dependencies, orientation=None):
		super().__init__(dependencies)
		self.name = name
		self.orientation = orientation

	def sampleGiven(self, value):
		return self

	def intersect(self, other, triedReversed=False):
		"""Get a `Region` representing the intersection of this one with another."""
		if triedReversed:
			return IntersectionRegion(self, other)
		else:
			return other.intersect(self, triedReversed=True)

	@staticmethod
	def uniformPointIn(region):
		"""Get a uniform `Distribution` over points in a `Region`."""
		return PointInRegionDistribution(region)

	def uniformPoint(self):
		"""Sample a uniformly-random point in this `Region`.

		Can only be called on fixed Regions with no random parameters.
		"""
		assert not needsSampling(self)
		return self.uniformPointInner()

	def uniformPointInner(self):
		"""Do the actual random sampling. Implemented by subclasses."""
		raise NotImplementedError()

	def containsPoint(self, point):
		"""Check if the `Region` contains a point. Implemented by subclasses."""
		raise NotImplementedError()

	def containsObject(self, obj):
		"""Check if the `Region` contains an :obj:`~scenic.core.object_types.Object`.

		The default implementation assumes the `Region` is convex; subclasses must
		override the method if this is not the case.
		"""
		for corner in obj.corners:
			if not self.containsPoint(corner):
				return False
		return True

	def __contains__(self, thing):
		"""Check if this `Region` contains an object or vector."""
		from scenic.core.object_types import Object
		if isinstance(thing, Object):
			return self.containsObject(thing)
		vec = toVector(thing, '"X in Y" with X not an Object or a vector')
		return self.containsPoint(vec)

	def getAABB(self):
		"""Axis-aligned bounding box for this `Region`. Implemented by some subclasses."""
		raise NotImplementedError()

	def orient(self, vec):
		"""Orient the given vector along the region's orientation, if any."""
		if self.orientation is None:
			return vec
		else:
			return OrientedVector(vec.x, vec.y, self.orientation[vec])

	def __str__(self):
		return f'<Region {self.name}>'

class AllRegion(Region):
	"""Region consisting of all space."""
	def intersect(self, other, triedReversed=False):
		return other

	def containsPoint(self, point):
		return True

	def containsObject(self, obj):
		return True

	def __eq__(self, other):
		return type(other) is AllRegion

	def __hash__(self):
		return hash(AllRegion)

class EmptyRegion(Region):
	"""Region containing no points."""
	def intersect(self, other, triedReversed=False):
		return self

	def uniformPointInner(self):
		raise RejectionException(f'sampling empty Region')

	def containsPoint(self, point):
		return False

	def containsObject(self, obj):
		return False

	def show(self, plt, style=None):
		pass

	def __eq__(self, other):
		return type(other) is EmptyRegion

	def __hash__(self):
		return hash(EmptyRegion)

everywhere = AllRegion('everywhere')
nowhere = EmptyRegion('nowhere')

class CircularRegion(Region):
	def __init__(self, center, radius, resolution=32):
		super().__init__('Circle', center, radius)
		self.center = center.toVector()
		self.radius = radius
		self.circumcircle = (self.center, self.radius)

		if not (needsSampling(self.center) or needsSampling(self.radius)):
			ctr = shapely.geometry.Point(self.center)
			self.polygon = ctr.buffer(self.radius, resolution=resolution)

	def sampleGiven(self, value):
		return CircularRegion(value[self.center], value[self.radius])

	def evaluateInner(self, context):
		center = valueInContext(self.center, context)
		radius = valueInContext(self.radius, context)
		return CircularRegion(center, radius)

	def containsPoint(self, point):
		point = point.toVector()
		return point.distanceTo(self.center) <= self.radius

	def uniformPointInner(self):
		x, y = self.center
		r = random.triangular(0, self.radius, self.radius)
		t = random.uniform(-math.pi, math.pi)
		pt = Vector(x + (r * cos(t)), y + (r * sin(t)))
		return self.orient(pt)

	def getAABB(self):
		x, y = self.center
		r = self.radius
		return ((x - r, y - r), (x + r, y + r))

	def isEquivalentTo(self, other):
		if type(other) is not CircularRegion:
			return False
		return (areEquivalent(other.center, self.center)
		        and areEquivalent(other.radius, self.radius))

	def __str__(self):
		return f'CircularRegion({self.center}, {self.radius})'

class SectorRegion(Region):
	def __init__(self, center, radius, heading, angle, resolution=32):
		super().__init__('Sector', center, radius, heading, angle)
		self.center = center.toVector()
		self.radius = radius
		self.heading = heading
		self.angle = angle
		r = (radius / 2) * cos(angle / 2)
		self.circumcircle = (self.center.offsetRadially(r, heading), r)

		if not any(needsSampling(x) for x in (self.center, radius, heading, angle)):
			ctr = shapely.geometry.Point(self.center)
			circle = ctr.buffer(self.radius, resolution=resolution)
			if angle >= math.tau - 0.001:
				self.polygon = circle
			else:
				mask = shapely.geometry.Polygon([
				    self.center,
				    self.center.offsetRadially(radius, heading + angle/2),
				    self.center.offsetRadially(2*radius, heading),
				    self.center.offsetRadially(radius, heading - angle/2)
				])
				self.polygon = circle & mask

	def sampleGiven(self, value):
		return SectorRegion(value[self.center], value[self.radius],
			value[self.heading], value[self.angle])

	def evaluateInner(self, context):
		center = valueInContext(self.center, context)
		radius = valueInContext(self.radius, context)
		heading = valueInContext(self.heading, context)
		angle = valueInContext(self.angle, context)
		return SectorRegion(center, radius, heading, angle)

	def containsPoint(self, point):
		point = point.toVector()
		if not pointIsInCone(tuple(point), tuple(self.center), self.heading, self.angle):
			return False
		return point.distanceTo(self.center) <= self.radius

	def uniformPointInner(self):
		x, y = self.center
		heading, angle, maxDist = self.heading, self.angle, self.radius
		r = random.triangular(0, maxDist, maxDist)
		ha = angle / 2.0
		t = random.uniform(-ha, ha) + (heading + (math.pi / 2))
		pt = Vector(x + (r * cos(t)), y + (r * sin(t)))
		return self.orient(pt)

	def isEquivalentTo(self, other):
		if type(other) is not SectorRegion:
			return False
		return (areEquivalent(other.center, self.center)
		        and areEquivalent(other.radius, self.radius)
		        and areEquivalent(other.heading, self.heading)
		        and areEquivalent(other.angle, self.angle))

	def __str__(self):
		return f'SectorRegion({self.center},{self.radius},{self.heading},{self.angle})'

class RectangularRegion(RotatedRectangle, Region):
	def __init__(self, position, heading, width, height):
		super().__init__('Rectangle', position, heading, width, height)
		self.position = position.toVector()
		self.heading = heading
		self.width = width
		self.height = height
		self.hw = hw = width / 2
		self.hh = hh = height / 2
		self.radius = hypot(hw, hh)		# circumcircle; for collision detection
		self.corners = tuple(position.offsetRotated(heading, Vector(*offset))
			for offset in ((hw, hh), (-hw, hh), (-hw, -hh), (hw, -hh)))
		self.circumcircle = (self.position, self.radius)

	def sampleGiven(self, value):
		return RectangularRegion(value[self.position], value[self.heading],
			value[self.width], value[self.height])

	def evaluateInner(self, context):
		position = valueInContext(self.position, context)
		heading = valueInContext(self.heading, context)
		width = valueInContext(self.width, context)
		height = valueInContext(self.height, context)
		return RectangularRegion(position, heading, width, height)

	def uniformPointInner(self):
		hw, hh = self.hw, self.hh
		rx = random.uniform(-hw, hw)
		ry = random.uniform(-hh, hh)
		pt = self.position.offsetRotated(self.heading, Vector(rx, ry))
		return self.orient(pt)

	def getAABB(self):
		x, y = zip(*self.corners)
		minx, maxx = findMinMax(x)
		miny, maxy = findMinMax(y)
		return ((minx, miny), (maxx, maxy))

	def isEquivalentTo(self, other):
		if type(other) is not RectangularRegion:
			return False
		return (areEquivalent(other.position, self.position)
		        and areEquivalent(other.heading, self.heading)
		        and areEquivalent(other.width, self.width)
		        and areEquivalent(other.height, self.height))

	def __str__(self):
		return f'RectangularRegion({self.position},{self.heading},{self.width},{self.height})'

class PolylineRegion(Region):
	"""Region given by one or more polylines (chain of line segments)"""
	def __init__(self, points=None, polyline=None, orientation=True):
		super().__init__('Polyline', orientation=orientation)
		if points is not None:
			points = tuple(points)
			if len(points) < 2:
				raise RuntimeError('tried to create PolylineRegion with < 2 points')
			self.points = points
			self.lineString = shapely.geometry.LineString(points)
		elif polyline is not None:
			if isinstance(polyline, shapely.geometry.LineString):
				if len(polyline.coords) < 2:
					raise RuntimeError('tried to create PolylineRegion with <2-point LineString')
			elif isinstance(polyline, shapely.geometry.MultiLineString):
				if len(polyline) == 0:
					raise RuntimeError('tried to create PolylineRegion from empty MultiLineString')
				for line in polyline:
					assert len(line.coords) >= 2
			else:
				raise RuntimeError('tried to create PolylineRegion from non-LineString')
			self.lineString = polyline
		else:
			raise RuntimeError('must specify points or polyline for PolylineRegion')
		if not self.lineString.is_valid:
			raise RuntimeError('tried to create PolylineRegion with '
			                   f'invalid LineString {self.lineString}')
		self.segments = self.segmentsOf(self.lineString)
		cumulativeLengths = []
		total = 0
		for p, q in self.segments:
			dx, dy = p[0] - q[0], p[1] - q[1]
			total += math.hypot(dx, dy)
			cumulativeLengths.append(total)
		self.cumulativeLengths = cumulativeLengths

	@classmethod
	def segmentsOf(cls, lineString):
		if isinstance(lineString, shapely.geometry.LineString):
			segments = []
			points = list(lineString.coords)
			if len(points) < 2:
				raise RuntimeError('LineString has fewer than 2 points')
			last = points[0]
			for point in points[1:]:
				segments.append((last, point))
				last = point
			return segments
		elif isinstance(lineString, shapely.geometry.MultiLineString):
			allSegments = []
			for line in lineString:
				allSegments.extend(cls.segmentsOf(line))
			return allSegments
		else:
			raise RuntimeError('called segmentsOf on non-linestring')

	def uniformPointInner(self):
		pointA, pointB = random.choices(self.segments,
		                                cum_weights=self.cumulativeLengths)[0]
		interpolation = random.random()
		x, y = averageVectors(pointA, pointB, weight=interpolation)
		if self.orientation is True:
			return OrientedVector(x, y, headingOfSegment(pointA, pointB))
		else:
			return self.orient(Vector(x, y))

	def intersect(self, other, triedReversed=False):
		poly = toPolygon(other)
		if poly is not None:
			intersection = self.lineString & poly
			if (intersection.is_empty or
			    not isinstance(intersection, (shapely.geometry.LineString,
			                                  shapely.geometry.MultiLineString))):
				# TODO handle points!
				return nowhere
			return PolylineRegion(polyline=intersection)
		return super().intersect(other, triedReversed)

	def containsPoint(self, point):
		return self.lineString.intersects(shapely.geometry.Point(point))

	def containsObject(self, obj):
		return False

	def getAABB(self):
		xmin, ymin, xmax, ymax = self.lineString.bounds
		return ((xmin, ymin), (xmax, ymax))

	def show(self, plt, style='r-'):
		for pointA, pointB in self.segments:
			plt.plot([pointA[0], pointB[0]], [pointA[1], pointB[1]], style)

	def __str__(self):
		return f'PolylineRegion({self.lineString})'

	def __eq__(self, other):
		if type(other) is not PolylineRegion:
			return NotImplemented
		return (other.lineString == self.lineString)

	@cached
	def __hash__(self):
		return hash(str(self.lineString))

class PolygonalRegion(Region):
	"""Region given by one or more polygons (possibly with holes)"""
	def __init__(self, points=None, polygon=None, orientation=None):
		super().__init__('Polygon', orientation=orientation)
		if polygon is None and points is None:
			raise RuntimeError('must specify points or polygon for PolygonalRegion')
		if polygon is None:
			points = tuple(points)
			if len(points) == 0:
				raise RuntimeError('tried to create PolygonalRegion from empty point list!')
			for point in points:
				if needsSampling(point):
					raise RuntimeError('only fixed PolygonalRegions are supported')
			self.points = points
			polygon = shapely.geometry.Polygon(points)

		if isinstance(polygon, shapely.geometry.Polygon):
			self.polygons = shapely.geometry.MultiPolygon([polygon])
		elif isinstance(polygon, shapely.geometry.MultiPolygon):
			self.polygons = polygon
		else:
			raise RuntimeError(f'tried to create PolygonalRegion from non-polygon {polygon}')
		if not self.polygons.is_valid:
			raise RuntimeError('tried to create PolygonalRegion with '
			                   f'invalid polygon {self.polygons}')

		if points is None and len(self.polygons) == 1 and len(self.polygons[0].interiors) == 0:
			self.points = tuple(self.polygons[0].exterior.coords[:-1])

		if self.polygons.is_empty:
			raise RuntimeError('tried to create empty PolygonalRegion')

		triangles = []
		for polygon in self.polygons:
			triangles.extend(triangulatePolygon(polygon))
		assert len(triangles) > 0, self.polygons
		self.trianglesAndBounds = tuple((tri, tri.bounds) for tri in triangles)
		areas = (triangle.area for triangle in triangles)
		self.cumulativeTriangleAreas = tuple(itertools.accumulate(areas))

	def uniformPointInner(self):
		triangle, bounds = random.choices(
			self.trianglesAndBounds,
			cum_weights=self.cumulativeTriangleAreas)[0]
		minx, miny, maxx, maxy = bounds
		# TODO improve?
		while True:
			x, y = random.uniform(minx, maxx), random.uniform(miny, maxy)
			if triangle.intersects(shapely.geometry.Point(x, y)):
				return self.orient(Vector(x, y))

	def intersect(self, other, triedReversed=False):
		poly = toPolygon(other)
		orientation = other.orientation if self.orientation is None else self.orientation
		if poly is not None:
			intersection = self.polygons & poly
			if intersection.is_empty:
				return nowhere
			elif isinstance(intersection, (shapely.geometry.Polygon,
			                             shapely.geometry.MultiPolygon)):
				return PolygonalRegion(polygon=intersection, orientation=orientation)
			elif isinstance(intersection, shapely.geometry.GeometryCollection):
				polys = []
				for geom in intersection:
					if isinstance(geom, shapely.geometry.Polygon):
						polys.append(geom)
				if len(polys) == 0:
					# TODO handle points, lines
					raise RuntimeError('unhandled type of polygon intersection')
				intersection = shapely.geometry.MultiPolygon(polys)
				return PolygonalRegion(polygon=intersection, orientation=orientation)
			else:
				# TODO handle points, lines
				raise RuntimeError('unhandled type of polygon intersection')
		return super().intersect(other, triedReversed)

	def union(self, other):
		poly = toPolygon(other)
		if not poly:
			raise RuntimeError(f'cannot take union of PolygonalRegion with {other}')
		union = polygonUnion((self.polygons, poly))
		return PolygonalRegion(polygon=union)

	def containsPoint(self, point):
		return self.polygons.intersects(shapely.geometry.Point(point))

	def containsObject(self, obj):
		objPoly = obj.polygon
		if objPoly is None:
			raise RuntimeError('tried to test containment of symbolic Object!')
		# TODO improve boundary handling?
		return self.polygons.contains(objPoly)

	def getAABB(self):
		xmin, xmax, ymin, ymax = self.polygons.bounds
		return ((xmin, ymin), (xmax, ymax))

	def show(self, plt, style='r-'):
		plotPolygon(self.polygons, plt, style=style)

	def __str__(self):
		return '<PolygonalRegion>'

	def __eq__(self, other):
		if type(other) is not PolygonalRegion:
			return NotImplemented
		return (other.polygons == self.polygons
		        and other.orientation == self.orientation)

	@cached
	def __hash__(self):
		# TODO better way to hash mutable Shapely geometries? (also for PolylineRegion)
		return hash((str(self.polygons), self.orientation))

class PointSetRegion(Region):
	"""Region consisting of a set of discrete points.

	No :obj:`~scenic.core.object_types.Object` can be contained in a `PointSetRegion`,
	since the latter is discrete. (This may not be true for subclasses, e.g.
	`GridRegion`.)

	Args:
		name (str): name for debugging
		points (iterable): set of points comprising the region
		kdtree (:obj:`scipy.spatial.KDTree`, optional): k-D tree for the points (one will
		  be computed if none is provided)
		orientation (:obj:`~scenic.core.vectors.VectorField`, optional): orientation for
		  the region
		tolerance (float, optional): distance tolerance for checking whether a point lies
		  in the region
	"""

	def __init__(self, name, points, kdTree=None, orientation=None, tolerance=1e-6):
		super().__init__(name, orientation=orientation)
		self.points = tuple(points)
		for point in self.points:
			if needsSampling(point):
				raise RuntimeError('only fixed PointSetRegions are supported')
		self.kdTree = scipy.spatial.cKDTree(self.points) if kdTree is None else kdTree
		self.orientation = orientation
		self.tolerance = tolerance

	def uniformPointInner(self):
		return self.orient(Vector(*random.choice(self.points)))

	def intersect(self, other, triedReversed=False):
		def sampler(intRegion):
			o = intRegion.regions[1]
			center, radius = o.circumcircle
			possibles = (Vector(*self.kdTree.data[i])
			             for i in self.kdTree.query_ball_point(center, radius))
			intersection = [p for p in possibles if o.containsPoint(p)]
			if len(intersection) == 0:
				raise RejectionException(f'empty intersection of Regions {self} and {o}')
			return self.orient(random.choice(intersection))
		return IntersectionRegion(self, other, sampler=sampler, orientation=self.orientation)

	def containsPoint(self, point):
		distance, location = self.kdTree.query(point)
		return (distance <= self.tolerance)

	def containsObject(self, obj):
		raise NotImplementedError()

	def __eq__(self, other):
		if type(other) is not PointSetRegion:
			return NotImplemented
		return (other.name == self.name
		        and other.points == self.points
		        and other.orientation == self.orientation)

	def __hash__(self):
		return hash((self.name, self.points, self.orientation))

class GridRegion(PointSetRegion):
	"""A Region given by an obstacle grid.

	A point is considered to be in a `GridRegion` if the nearest grid point is
	not an obstacle.

	Args:
		name (str): name for debugging
		grid: 2D list, tuple, or NumPy array of 0s and 1s, where 1 indicates an obstacle
		  and 0 indicates free space
		Ax (float): spacing between grid points along X axis
		Ay (float): spacing between grid points along Y axis
		Bx (float): X coordinate of leftmost grid column
		By (float): Y coordinate of lowest grid row
		orientation (:obj:`~scenic.core.vectors.VectorField`, optional): orientation of region
	"""
	def __init__(self, name, grid, Ax, Ay, Bx, By, orientation=None):
		self.grid = numpy.array(grid)
		self.sizeY, self.sizeX = self.grid.shape
		self.Ax, self.Ay = Ax, Ay
		self.Bx, self.By = Bx, By
		y, x = numpy.where(self.grid == 0)
		points = [self.gridToPoint(point) for point in zip(x, y)]
		super().__init__(name, points, orientation=orientation)

	def gridToPoint(self, gp):
		x, y = gp
		return ((self.Ax * x) + self.Bx, (self.Ay * y) + self.By)

	def pointToGrid(self, point):
		x, y = point
		x = (x - self.Bx) / self.Ax
		y = (y - self.By) / self.Ay
		nx = int(round(x))
		if nx < 0 or nx >= self.sizeX:
			return None
		ny = int(round(y))
		if ny < 0 or ny >= self.sizeY:
			return None
		return (nx, ny)

	def containsPoint(self, point):
		gp = self.pointToGrid(point)
		if gp is None:
			return False
		x, y = gp
		return (self.grid[y, x] == 0)

	def containsObject(self, obj):
		# TODO improve this procedure!
		# Fast check
		for c in obj.corners:
			if not self.containsPoint(c):
				return False
		# Slow check
		gps = [self.pointToGrid(corner) for corner in obj.corners]
		x, y = zip(*gps)
		minx, maxx = findMinMax(x)
		miny, maxy = findMinMax(y)
		for x in range(minx, maxx+1):
			for y in range(miny, maxy+1):
				p = self.gridToPoint((x, y))
				if self.grid[y, x] == 1 and obj.containsPoint(p):
					return False
		return True

class IntersectionRegion(Region):
	def __init__(self, *regions, orientation=None, sampler=None):
		self.regions = tuple(regions)
		if len(self.regions) < 2:
			raise RuntimeError('tried to take intersection of fewer than 2 regions')
		super().__init__('Intersection', *self.regions, orientation=orientation)
		if sampler is None:
			sampler = self.genericSampler
		self.sampler = sampler

	def sampleGiven(self, value):
		regs = [value[reg] for reg in self.regions]
		# Now that regions have been sampled, attempt intersection again in the hopes
		# there is a specialized sampler to handle it (unless we already have one)
		if self.sampler is self.genericSampler:
			failed = False
			intersection = regs[0]
			for region in regs[1:]:
				intersection = intersection.intersect(region)
				if isinstance(intersection, IntersectionRegion):
					failed = True
					break
			if not failed:
				intersection.orientation = value[self.orientation]
				return intersection
		return IntersectionRegion(*regs, orientation=value[self.orientation],
		                          sampler=self.sampler)

	def evaluateInner(self, context):
		regs = (valueInContext(reg, context) for reg in self.regions)
		orientation = valueInContext(self.orientation, context)
		return IntersectionRegion(*regs, orientation=orientation, sampler=self.sampler)

	def containsPoint(self, point):
		return all(region.containsPoint(point) for region in self.regions)

	def uniformPointInner(self):
		return self.orient(self.sampler(self))

	@staticmethod
	def genericSampler(intersection):
		regs = intersection.regions
		point = regs[0].uniformPointInner()
		for region in regs[1:]:
			if not region.containsPoint(point):
				raise RejectionException(
				    f'sampling intersection of Regions {regs[0]} and {region}')
		return point

	def isEquivalentTo(self, other):
		if type(other) is not IntersectionRegion:
			return False
		return (areEquivalent(set(other.regions), set(self.regions))
		        and other.orientation == self.orientation)

	def __str__(self):
		return f'IntersectionRegion({self.regions})'




### Top-level functionality of the scenic package as a script:
### load a scenario and generate scenes in an infinite loop.

import sys
import time
import argparse
import random

import scenic.syntax.translator as translator

parser = argparse.ArgumentParser(prog='scenic',
                                 usage='scenic [-h] [options] scenario',
                                 description='Interactively sample from a Scenic scenario.')

# Options
parser.add_argument('-d', '--delay', help='loop automatically with this delay (in seconds)',
                    type=float)
parser.add_argument('-z', '--zoom', help='zoom expansion factor', type=float, default=2)
parser.add_argument('-s', '--seed', help='random seed', type=int)
parser.add_argument('-v', '--verbosity', help='verbosity level (default 1)',
                    type=int, choices=(0, 1, 2, 3), default=1)

# Debugging options
debugOpts = parser.add_argument_group('debugging options')
debugOpts.add_argument('-b', '--full-backtrace', help='show full internal backtraces',
                       action='store_true')
debugOpts.add_argument('--dump-python', help='dump translated Python', action='store_true')
debugOpts.add_argument('--dump-ast', help='dump final AST', action='store_true')
debugOpts.add_argument('--no-pruning', help='disable pruning', action='store_true')
debugOpts.add_argument('--gather-stats', help='collect statistics over this many scenes',
                       type=int, metavar='N')

# Positional arguments
parser.add_argument('scenario', help='a Scenic file to run')

# Parse arguments and set up configuration
args = parser.parse_args()
delay = args.delay
translator.showInternalBacktrace = args.full_backtrace
translator.dumpTranslatedPython = args.dump_python
translator.dumpFinalAST = args.dump_ast
translator.verbosity = args.verbosity
translator.usePruning = not args.no_pruning
if args.seed is not None:
    print(f'Using random seed = {args.seed}')
    random.seed(args.seed)

# Load scenario from file
print('Beginning scenario construction...')
startTime = time.time()
scenario = translator.scenarioFromFile(args.scenario)
totalTime = time.time() - startTime
print(f'Scenario constructed in {totalTime:.2f} seconds.')

def generateScene():
    startTime = time.time()
    scene, iterations = scenario.generate(verbosity=args.verbosity)
    if args.verbosity >= 1:
        totalTime = time.time() - startTime
        print(f'  Generated scene in {iterations} iterations, {totalTime:.4g} seconds.')
    return scene, iterations

if args.gather_stats is None:   # Generate scenes interactively until killed
    import matplotlib.pyplot as plt
    while True:
        scene, _ = generateScene()
        if delay is None:
            scene.show(zoom=args.zoom)
        else:
            scene.show(zoom=args.zoom, block=False)
            plt.pause(delay)
            plt.clf()
else:   # Gather statistics over the specified number of scenes
    its = []
    startTime = time.time()
    while len(its) < args.gather_stats:
        scene, iterations = generateScene()
        its.append(iterations)
    totalTime = time.time() - startTime
    count = len(its)
    print(f'Sampled {len(its)} scenes in {totalTime:.2f} seconds.')
    print(f'Average iterations/scene: {sum(its)/count}')
    print(f'Average time/scene: {totalTime/count:.2f} seconds.')



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import json
import numpy as np
from numpy.lib.stride_tricks import as_strided

def obj_dict(obj):
    return obj.__dict__

class Formal_Config():
    def __init__(self, location=None, time=None, weather=None, vehicles=None, view_heading=None):
        self.location = location
        self.time = time
        self.weather = weather
        self.vehicles = vehicles
        self.view_heading = view_heading

class Formal_Configs():
    def __init__(self, cfgs=None):
        self.cfgs = cfgs
    
    def to_json(self):
        return json.dumps({"formal_configs" : self.cfgs}, default=obj_dict)

class Vehicle():
    def __init__(self, model=None, color=None, location_offset=None, heading=None):
        self.model = model
        self.color = color
        self.location_offset = location_offset
        self.heading = heading

    def __repr__(self):
        return str(self.__dict__)

class Scenario:
    def __init__(self, location=None, time=None, weather=None, vehicle=None, drivingMode=None):
        self.location = location #[x,y]
        self.time = time #[hour, minute]
        self.weather = weather #string
        self.vehicle = vehicle #string
        self.drivingMode = drivingMode #[drivingMode, setSpeed]

class Dataset:
    def __init__(self, rate=None, frame=None, vehicles=None, peds=None, trafficSigns=None, direction=None, reward=None, 
            throttle=None, brake=None, steering=None, speed=None, yawRate=None, drivingMode=None, location=None, time=None):
        self.rate = rate #Hz
        self.frame = frame #[width, height]
        self.vehicles = vehicles #boolean
        self.peds = peds #boolean
        self.trafficSigns = trafficSigns #boolean
        self.direction = direction #[x,y,z]
        self.reward = reward #[id, p1, p2]
        self.throttle = throttle #boolean
        self.brake = brake #boolean
        self.steering = steering #boolean
        self.speed = speed #boolean
        self.yawRate = yawRate #boolean
        self.drivingMode = drivingMode #boolean
        self.location = location #boolean
        self.time = time #boolean

class Start:
    def __init__(self, scenario=None, dataset=None):
        self.scenario = scenario
        self.dataset = dataset

    def to_json(self):
        _scenario = None
        _dataset = None

        if (self.scenario != None):
            _scenario = self.scenario.__dict__

        if (self.dataset != None):
            _dataset = self.dataset.__dict__            

        return json.dumps({'start':{'scenario': _scenario, 'dataset': _dataset}})


class Config:
    def __init__(self, scenario=None, dataset=None):
        self.scenario = scenario
        self.dataset = dataset

    def to_json(self):
        _scenario = None
        _dataset = None

        if (self.scenario != None):
            _scenario = self.scenario.__dict__

        if (self.dataset != None):
            _dataset = self.dataset.__dict__            

        return json.dumps({'config':{'scenario': _scenario, 'dataset': _dataset}})

class Stop:
    def to_json(self):
        return json.dumps({'stop':None}) #super dummy

class Commands:
    def __init__(self, throttle=None, brake=None, steering=None):
        self.throttle = throttle #float (0,1)
        self.brake = brake #float (0,1)
        self.steering = steering #float (-1,1)

    def to_json(self):
        return json.dumps({'commands':self.__dict__})
        
def frame2numpy(frame, frameSize):
    buff = np.fromstring(frame, dtype='uint8')
    # Scanlines are aligned to 4 bytes in Windows bitmaps
    strideWidth = int((frameSize[0] * 3 + 3) / 4) * 4
    # Return a copy because custom strides are not supported by OpenCV.
    return as_strided(buff, strides=(strideWidth, 3, 1), shape=(frameSize[1], frameSize[0], 3)).copy()




"""Scenic world model for the X-Plane flight simulator.

See the `VerifAI distribution`_ for examples of how to use Scenic with X-Plane.

.. raw:: html

   <h2>Submodules</h2>

.. autosummary::
   :toctree: _autosummary

   model

.. _VerifAI distribution: https://github.com/BerkeleyLearnVerify/VerifAI
"""




# stub to allow changing the map without having to alter gta_model.sc

import os

mapPath = 'map.npz'

def setLocalMap(module, relpath):
    global mapPath
    base = os.path.dirname(module)
    mapPath = os.path.join(base, relpath)



"""Scenic's core types and associated support code.

.. raw:: html

   <h2>Submodules</h2>

.. autosummary::
   :toctree: _autosummary

   distributions
   external_params
   geometry
   lazy_eval
   object_types
   pruning
   regions
   scenarios
   specifiers
   type_support
   utils
   vectors
   workspaces
"""




"""Assorted utility functions and common exceptions."""

import functools
import math

sqrt2 = math.sqrt(2)

def cached(oldMethod):
    """Decorator for making a method with no arguments cache its result"""
    storageName = f'_cached_{oldMethod.__name__}'
    @functools.wraps(oldMethod)
    def newMethod(self):
        try:
            # Use __getattribute__ for direct lookup in case self is a Distribution
            return self.__getattribute__(storageName)
        except AttributeError:
            value = oldMethod(self)
            setattr(self, storageName, value)
            return value
    return newMethod

def argsToString(args):
    names = (f'{a[0]}={a[1]}' if isinstance(a, tuple) else str(a) for a in args)
    joinedArgs = ', '.join(names)
    return f'({joinedArgs})'

def areEquivalent(a, b):
    """Whether two objects are equivalent, i.e. have the same properties.

    This is only used for debugging, e.g. to check that a Distribution is the
    same before and after pickling. We don't want to define __eq__ for such
    objects since for example two values sampled with the same distribution are
    equivalent but not semantically identical: the code::

        X = (0, 1)
        Y = (0, 1)

    does not make X and Y always have equal values!"""
    if isinstance(a, (list, tuple)) and isinstance(b, (list, tuple)):
        if len(a) != len(b):
            return False
        for x, y in zip(a, b):
            if not areEquivalent(x, y):
                return False
        return True
    elif isinstance(a, (set, frozenset)) and isinstance(b, (set, frozenset)):
        if len(a) != len(b):
            return False
        mb = set(b)
        for x in a:
            found = False
            for y in mb:
                if areEquivalent(x, y):
                    mb.remove(y)
                    found = True
                    break
            if not found:
                return False
        return True
    elif isinstance(a, dict) and isinstance(b, dict):
        if len(a) != len(b):
            return False
        for x, v in a.items():
            found = False
            for y, w in b.items():
                if areEquivalent(x, y) and areEquivalent(v, w):
                    del b[y]
                    found = True
                    break
            if not found:
                return False
        return True
    elif hasattr(a, 'isEquivalentTo'):
        return a.isEquivalentTo(b)
    elif hasattr(b, 'isEquivalentTo'):
        return b.isEquivalentTo(a)
    else:
        return a == b

class ParseError(Exception):
    """An error produced by attempting to parse an invalid Scenic program."""
    pass

class RuntimeParseError(ParseError):
    """A Scenic parse error generated during execution of the translated Python."""
    pass

class InvalidScenarioError(Exception):
    """Error raised for syntactically-valid but otherwise problematic Scenic programs."""
    pass

class InconsistentScenarioError(InvalidScenarioError):
    """Error for scenarios with inconsistent requirements."""
    def __init__(self, line, message):
        self.lineno = line
        super().__init__('Inconsistent requirement on line ' + str(line) + ': ' + message)




"""Common Webots interface."""

import math

import numpy as np

from scenic.core.geometry import normalizeAngle

def webotsToScenicPosition(pos):
    """Convert Webots positions to Scenic positions."""
    x, y, z = pos
    return (x, -z)

def scenicToWebotsPosition(pos, y=0):
    x, z = pos
    return [x, y, -z]

def webotsToScenicRotation(rot, tolerance2D=0):
    axis = np.array(rot[:3])
    angle = rot[3]
    if np.linalg.norm(axis - (0, 1, 0)) > tolerance2D:
        return None
    return normalizeAngle(angle + math.pi)

def scenicToWebotsRotation(heading):
    return [0, 1, 0, heading - math.pi]




"""Pruning parts of the sample space which violate requirements."""

import math
import time
import shapely.geometry
import shapely.geos

from scenic.core.distributions import (Samplable, MethodDistribution, OperatorDistribution,
                                       needsSampling, supportInterval, underlyingFunction)
from scenic.core.object_types import Point, Object
from scenic.core.geometry import normalizeAngle, polygonUnion, plotPolygon
from scenic.core.vectors import VectorField, PolygonalVectorField, VectorMethodDistribution
from scenic.core.workspaces import Workspace
from scenic.syntax.relations import RelativeHeadingRelation, DistanceRelation
from scenic.core.utils import InvalidScenarioError
import scenic.core.regions as regions

### Utilities

def currentPropValue(obj, prop):
    """Get the current value of an object's property, taking into account prior pruning."""
    value = getattr(obj, prop)
    return value._conditioned if isinstance(value, Samplable) else value

def isMethodCall(thing, method):
    """Match calls to a given method, taking into account distribution decorators."""
    if not isinstance(thing, (MethodDistribution, VectorMethodDistribution)):
        return False
    return thing.method is underlyingFunction(method)

def matchInRegion(position):
    """Match uniform samples from a Region, returning the Region if any."""
    position = position.toVector()
    if isinstance(position, regions.PointInRegionDistribution):
        reg = position.region
        if isinstance(reg, Workspace):
            reg = reg.region
        return reg
    return None

def matchPolygonalField(heading, position):
    """Match headings defined by a PolygonalVectorField at the given position.

    Matches headings exactly equal to a PolygonalVectorField, or offset by a
    bounded disturbance. Returns a triplet consisting of the matched field if
    any, together with lower/upper bounds on the disturbance.
    """
    if (isMethodCall(heading, VectorField.__getitem__)
        and isinstance(heading.object, PolygonalVectorField)
        and heading.arguments == (position.toVector(),)):
        return heading.object, 0, 0
    elif isinstance(heading, OperatorDistribution) and heading.operator in ('__add__', '__radd__'):
        field, lower, upper = matchPolygonalField(heading.object, position)
        if field is not None:
            assert len(heading.operands) == 1
            offset = heading.operands[0]
            ol, oh = supportInterval(offset)
            if ol is not None and oh is not None:
                return field, lower + ol, upper + oh
    return None, 0, 0

### Pruning procedures

def prune(scenario, verbosity=1):
    """Prune a Scenario, removing infeasible parts of the space.

    This function directly modifies the Distributions used in the Scenario,
    but leaves the conditional distribution under the scenario's requirements
    unchanged.
    """
    if verbosity >= 1:
        print('  Pruning scenario...')
        startTime = time.time()

    pruneContainment(scenario, verbosity)
    pruneRelativeHeading(scenario, verbosity)

    if verbosity >= 1:
        totalTime = time.time() - startTime
        print(f'  Pruned scenario in {totalTime:.4g} seconds.')

## Pruning based on containment

def pruneContainment(scenario, verbosity):
    """Prune based on the requirement that individual Objects fit within their container.

    Specifically, if O is positioned uniformly in region B and has container C, then we
    can instead pick a position uniformly in their intersection. If we can also lower
    bound the radius of O, then we can first erode C by that distance.
    """
    for obj in scenario.objects:
        base = matchInRegion(obj.position)
        if base is None:                    # match objects positioned uniformly in a Region
            continue
        if isinstance(base, regions.EmptyRegion):
            raise InvalidScenarioError(f'Object {obj} placed in empty region')
        basePoly = regions.toPolygon(base)
        if basePoly is None:                # to prune, the Region must be polygonal
            continue
        if basePoly.is_empty:
            raise InvalidScenarioError(f'Object {obj} placed in empty region')
        container = scenario.containerOfObject(obj)
        containerPoly = regions.toPolygon(container)
        if containerPoly is None:           # the object's container must also be polygonal
            return None
        minRadius, _ = supportInterval(obj.inradius)
        if minRadius is not None:           # if we can lower bound the radius, erode the container
            containerPoly = containerPoly.buffer(-minRadius)
        elif base is container:
            continue
        newBasePoly = basePoly & containerPoly      # restrict the base Region to the container
        if newBasePoly.is_empty:
            raise InvalidScenarioError(f'Object {obj} does not fit in container')
        if verbosity >= 1:
            if basePoly.area > 0:
                ratio = newBasePoly.area / basePoly.area
            else:
                ratio = newBasePoly.length / basePoly.length
            percent = 100 * (1.0 - ratio)
            print(f'    Region containment constraint pruned {percent:.1f}% of space.')
        newBase = regions.regionFromShapelyObject(newBasePoly, orientation=base.orientation)
        newPos = regions.Region.uniformPointIn(newBase)
        obj.position.conditionTo(newPos)

## Pruning based on orientation

def pruneRelativeHeading(scenario, verbosity):
    """Prune based on requirements bounding the relative heading of an Object.

    Specifically, if an object O is:

        * positioned uniformly within a polygonal region B;
        * aligned to a polygonal vector field F (up to a bounded offset);

    and another object O' is:

        * aligned to a polygonal vector field F' (up to a bounded offset);
        * at most some finite maximum distance from O;
        * required to have relative heading within a bounded offset of that of O;

    then we can instead position O uniformly in the subset of B intersecting the cells
    of F which satisfy the relative heading requirements w.r.t. some cell of F' which
    is within the distance bound.
    """
    # Check which objects are (approximately) aligned to polygonal vector fields
    fields = {}
    for obj in scenario.objects:
        field, offsetL, offsetR = matchPolygonalField(obj.heading, obj.position)
        if field is not None:
            fields[obj] = (field, offsetL, offsetR)
    # Check for relative heading relations among such objects
    for obj, (field, offsetL, offsetR) in fields.items():
        position = currentPropValue(obj, 'position')
        base = matchInRegion(position)
        if base is None:        # obj must be positioned uniformly in a Region
            continue
        basePoly = regions.toPolygon(base)
        if basePoly is None:    # the Region must be polygonal
            continue
        newBasePoly = basePoly
        for rel in obj._relations:
            if isinstance(rel, RelativeHeadingRelation) and rel.target in fields:
                tField, tOffsetL, tOffsetR = fields[rel.target]
                maxDist = maxDistanceBetween(scenario, obj, rel.target)
                if maxDist == float('inf'):     # the distance between the objects must be bounded
                    continue
                feasible = feasibleRHPolygon(field, offsetL, offsetR,
                                             tField, tOffsetL, tOffsetR,
                                             rel.lower, rel.upper, maxDist)
                if feasible is None:    # the RH bounds may be too weak to restrict the space
                    continue
                try:
                    pruned = newBasePoly & feasible
                except shapely.geos.TopologicalError:   # TODO how can we prevent these??
                    pruned = newBasePoly & feasible.buffer(0.1, cap_style=2)
                if verbosity >= 1:
                    percent = 100 * (1.0 - (pruned.area / newBasePoly.area))
                    print(f'    Relative heading constraint pruned {percent:.1f}% of space.')
                newBasePoly = pruned
        if newBasePoly is not basePoly:
            newBase = regions.PolygonalRegion(polygon=newBasePoly,
                                              orientation=base.orientation)
            newPos = regions.Region.uniformPointIn(newBase)
            obj.position.conditionTo(newPos)

def maxDistanceBetween(scenario, obj, target):
    """Upper bound the distance between the given Objects."""
    # If one of the objects is the ego, use visibility requirements
    ego = scenario.egoObject
    if obj is ego and target.requireVisible:
        visDist = visibilityBound(ego, target)
    elif target is ego and obj.requireVisible:
        visDist = visibilityBound(ego, obj)
    else:
        visDist = float('inf')

    # Check for any distance bounds implied by user-specified requirements
    reqDist = float('inf')
    for rel in obj._relations:
        if isinstance(rel, DistanceRelation) and rel.target is target:
            if rel.upper < reqDist:
                reqDist = rel.upper

    return min(visDist, reqDist)

def visibilityBound(obj, target):
    """Upper bound the distance from an Object to another it can see."""
    # Upper bound on visible distance is a sum of several terms:
    # 1. obj.visibleDistance
    _, maxVisibleDistance = supportInterval(obj.visibleDistance)
    if maxVisibleDistance is None:
        return None
    # 2. distance from obj's center to its camera
    _, maxCameraX = supportInterval(obj.cameraOffset.x)
    _, maxCameraY = supportInterval(obj.cameraOffset.y)
    if maxCameraX is None or maxCameraY is None:
        return None
    maxVisibleDistance += math.hypot(maxCameraX, maxCameraY)
    # 3. radius of target
    _, maxRadius = supportInterval(target.radius)
    if maxRadius is None:
        return None
    maxVisibleDistance += maxRadius
    return maxVisibleDistance

def feasibleRHPolygon(field, offsetL, offsetR,
                      tField, tOffsetL, tOffsetR,
                      lowerBound, upperBound, maxDist):
    """Find where objects aligned to the given fields can satisfy the given RH bounds."""
    if (offsetR - offsetL >= math.tau
        or tOffsetR - tOffsetL >= math.tau
        or upperBound - lowerBound >= math.tau):
        return None
    polygons = []
    expanded = [(poly.buffer(maxDist), heading) for poly, heading in tField.cells]
    for baseCell, baseHeading in field.cells:   # TODO skip cells not contained in base region?
        for expandedTargetCell, targetHeading in expanded:
            lower, upper = relativeHeadingRange(baseHeading, offsetL, offsetR,
                                                targetHeading, tOffsetL, tOffsetR)
            if (upper >= lowerBound and lower <= upperBound):   # RH intervals overlap
                intersection = baseCell & expandedTargetCell
                if not intersection.is_empty:
                    assert isinstance(intersection, shapely.geometry.Polygon), intersection
                    polygons.append(intersection)
    return polygonUnion(polygons)

def relativeHeadingRange(baseHeading, offsetL, offsetR,
                         targetHeading, tOffsetL, tOffsetR):
    """Lower/upper bound the possible RH between two headings with bounded disturbances."""
    if baseHeading is None or targetHeading is None:    # heading may not be constant within cell
        return -math.pi, math.pi
    lower = normalizeAngle(baseHeading + offsetL)
    upper = normalizeAngle(baseHeading + offsetR)
    points = [lower, upper]
    if upper < lower:
        points.extend((math.pi, -math.pi))
    tLower = normalizeAngle(targetHeading + tOffsetL)
    tUpper = normalizeAngle(targetHeading + tOffsetR)
    tPoints = [tLower, tUpper]
    if tUpper < tLower:
        tPoints.extend((math.pi, -math.pi))
    rhs = [tp - p for tp in tPoints for p in points]    # TODO improve
    return min(rhs), max(rhs)



"""Parser for WBT files using ANTLR.

The ANTLR parser itself, consisting of the *WBTLexer.py*, *WBTParser.py*, and *WBTVisitor.py* files, is autogenerated from *WBT.g4*.
"""

import sys
import io
import time

import numpy as np

import antlr4
from antlr4.error.ErrorListener import ErrorListener
from scenic.simulators.webots.WBTLexer import WBTLexer
from scenic.simulators.webots.WBTParser import WBTParser
from scenic.simulators.webots.WBTVisitor import WBTVisitor

### Setup

## Classes representing various types of VRML nodes 

class Node:
	"""A generic VRML node"""
	def __init__(self, nodeType, attrs):
		self.nodeType = nodeType
		self.attrs = attrs

## Classes for parsing with ANTLR

# Error handling

class ErrorReporter(ErrorListener):
	"""ANTLR listener for reporting parse errors"""
	def syntaxError(self, recognizer, offendingSymbol, line, column, msg, e):
		raise Exception(f'Parse error at {line}:{column}: {msg}')

# Walking the parse tree and extracting values

class Evaluator(WBTVisitor):
	"""Constructs an object representing the given value from the parse tree"""
	def __init__(self, nodeClasses):
		super().__init__()
		# Dictionary indicating which types of nodes have specialized classes
		self.nodeClasses = nodeClasses

	def visitNode(self, node):
		nodeType = node.Identifier().getText()
		attrs = self.visitNodeBody(node.nodeBody())
		if nodeType in self.nodeClasses:
			nodeClass = self.nodeClasses[nodeType]
			nodeObj = nodeClass(attrs)
		else:
			nodeObj = Node(nodeType, attrs)
		return nodeObj

	def visitNodeBody(self, body):
		return {attr.Identifier().getText(): self.visitValue(attr.value())
			for attr in body.attribute()}

	def visitValue(self, val):
		return self.visit(val.getChild(0))

	def visitVector(self, vec):
		return np.array([float(n.getText()) for n in vec.Number()])

	def visitVectorWithNewlines(self, vec):
		return self.visitVector(vec)

	def visitString(self, string):
		return string.getText()[1:-1]	# strip off quotes

	def visitArray(self, array):
		return [self.visit(val) for val in array.children
			if not isinstance(val, antlr4.tree.Tree.TerminalNode)]

	def visitBoolean(self, boolean):
		return boolean.getText() == 'TRUE'

def parse(path):
	"""Parse a world from a WBT file"""
	# Open file
	charStream = antlr4.FileStream(path)

	# Set up lexer and parser
	errorListener = ErrorReporter()
	lexer = WBTLexer(charStream)
	lexer.addErrorListener(errorListener)
	tokenStream = antlr4.CommonTokenStream(lexer)
	parser = WBTParser(tokenStream)
	parser.addErrorListener(errorListener)

	# Parse file
	return parser.world()

def findNodeTypesIn(types, world, nodeClasses={}):
	"""Find all nodes of the given types in a world"""
	evaluator = Evaluator(nodeClasses)
	types = tuple(types)
	instances = {ty: [] for ty in types}
	for node in world.children:
		if isinstance(node, WBTParser.DefnContext):
			node = node.node()
		nodeType = node.Identifier().getText()
		if nodeType in instances:
			val = evaluator.visitNode(node)
			instances[nodeType].append(val)
	return tuple(instances[ty] for ty in types)



"""Scenario and scene objects."""

import random
import time

from scenic.core.distributions import Samplable, RejectionException, needsSampling
from scenic.core.lazy_eval import needsLazyEvaluation
from scenic.core.external_params import ExternalSampler
from scenic.core.workspaces import Workspace
from scenic.core.vectors import Vector
from scenic.core.utils import areEquivalent, InvalidScenarioError

class Scene:
	"""A scene generated from a Scenic scenario.

	Attributes:
		objects (tuple(:obj:`~scenic.core.object_types.Object`)): All objects in the
		  scene. The ``ego`` object is first.
		egoObject (:obj:`~scenic.core.object_types.Object`): The ``ego`` object.
		params (dict): Dictionary mapping the name of each global parameter to its value.
		workspace (:obj:`~scenic.core.workspaces.Workspace`): Workspace for the scenario.
	"""
	def __init__(self, workspace, objects, egoObject, params):
		self.workspace = workspace
		self.objects = tuple(objects)
		self.egoObject = egoObject
		self.params = params

	def show(self, zoom=None, block=True):
		"""Render a schematic of the scene for debugging."""
		import matplotlib.pyplot as plt
		# display map
		self.workspace.show(plt)
		# draw objects
		for obj in self.objects:
			obj.show(self.workspace, plt, highlight=(obj is self.egoObject))
		# zoom in if requested
		if zoom != None:
			self.workspace.zoomAround(plt, self.objects, expansion=zoom)
		plt.show(block=block)

class Scenario:
	"""A compiled Scenic scenario, from which scenes can be sampled."""
	def __init__(self, workspace,
	             objects, egoObject,
	             params, externalParams,
	             requirements, requirementDeps):
		if workspace is None:
			workspace = Workspace()		# default empty workspace
		self.workspace = workspace
		ordered = []
		for obj in objects:
			ordered.append(obj)
			if obj is egoObject:	# make ego the first object
				ordered[0], ordered[-1] = ordered[-1], ordered[0]
		assert ordered[0] is egoObject
		self.objects = tuple(ordered)
		self.egoObject = egoObject
		self.params = dict(params)
		self.externalParams = tuple(externalParams)
		self.requirements = tuple(requirements)
		self.externalSampler = ExternalSampler.forParameters(self.externalParams, self.params)
		# dependencies must use fixed order for reproducibility
		paramDeps = tuple(p for p in self.params.values() if isinstance(p, Samplable))
		self.dependencies = self.objects + paramDeps + tuple(requirementDeps)
		self.validate()

	def isEquivalentTo(self, other):
		if type(other) is not Scenario:
			return False
		return (areEquivalent(other.workspace, self.workspace)
		    and areEquivalent(other.objects, self.objects)
		    and areEquivalent(other.params, self.params)
		    and areEquivalent(other.externalParams, self.externalParams)
		    and areEquivalent(other.requirements, self.requirements)
		    and other.externalSampler == self.externalSampler)

	def containerOfObject(self, obj):
		if hasattr(obj, 'regionContainedIn') and obj.regionContainedIn is not None:
			return obj.regionContainedIn
		else:
			return self.workspace.region

	def validate(self):
		"""Make some simple static checks for inconsistent built-in requirements."""
		objects = self.objects
		staticVisibility = not needsSampling(self.egoObject.visibleRegion)
		staticBounds = [self.hasStaticBounds(obj) for obj in objects]
		for i in range(len(objects)):
			oi = objects[i]
			# skip objects with unknown positions or bounding boxes
			if not staticBounds[i]:
				continue
			# Require object to be contained in the workspace/valid region
			container = self.containerOfObject(oi)
			if not needsSampling(container) and not container.containsObject(oi):
				raise InvalidScenarioError(f'Object at {oi.position} does not fit in container')
			# Require object to be visible from the ego object
			if staticVisibility and oi.requireVisible is True and oi is not self.egoObject:
				if not self.egoObject.canSee(oi):
					raise InvalidScenarioError(f'Object at {oi.position} is not visible from ego')
			# Require object to not intersect another object
			for j in range(i):
				oj = objects[j]
				if not staticBounds[j]:
					continue
				if oi.intersects(oj):
					raise InvalidScenarioError(f'Object at {oi.position} intersects'
					                           f' object at {oj.position}')

	def hasStaticBounds(self, obj):
		if needsSampling(obj.position):
			return False
		if any(needsSampling(corner) for corner in obj.corners):
			return False
		return True

	def generate(self, maxIterations=2000, verbosity=0, feedback=None):
		"""Sample a `Scene` from this scenario.

		Args:
			maxIterations (int): Maximum number of rejection sampling iterations.
			verbosity (int): Verbosity level.
			feedback (float): Feedback to pass to external samplers doing active sampling.
				See :mod:`scenic.core.external_params`.

		Returns:
			A pair with the sampled `Scene` and the number of iterations used.

		Raises:
			`RejectionException`: if no valid sample is found in **maxIterations** iterations.
		"""
		objects = self.objects

		# choose which custom requirements will be enforced for this sample
		activeReqs = [req for req, prob in self.requirements if random.random() <= prob]

		# do rejection sampling until requirements are satisfied
		rejection = True
		iterations = 0
		while rejection is not None:
			if iterations > 0:	# rejected the last sample
				if verbosity >= 2:
					print(f'  Rejected sample {iterations} because of: {rejection}')
				if self.externalSampler is not None:
					feedback = self.externalSampler.rejectionFeedback
			if iterations >= maxIterations:
				raise RejectionException(f'failed to generate scenario in {iterations} iterations')
			iterations += 1
			try:
				if self.externalSampler is not None:
					self.externalSampler.sample(feedback)
				sample = Samplable.sampleAll(self.dependencies)
			except RejectionException as e:
				rejection = e
				continue
			rejection = None
			ego = sample[self.egoObject]
			# Normalize types of some built-in properties
			for obj in objects:
				sampledObj = sample[obj]
				assert not needsSampling(sampledObj)
				assert isinstance(sampledObj.position, Vector)
				sampledObj.heading = float(sampledObj.heading)
			# Check built-in requirements
			for i in range(len(objects)):
				vi = sample[objects[i]]
				# Require object to be contained in the workspace/valid region
				container = self.containerOfObject(vi)
				if not container.containsObject(vi):
					rejection = 'object containment'
					break
				# Require object to be visible from the ego object
				if vi.requireVisible and vi is not ego and not ego.canSee(vi):
					rejection = 'object visibility'
					break
				# Require object to not intersect another object
				for j in range(i):
					vj = sample[objects[j]]
					if vi.intersects(vj):
						rejection = 'object intersection'
						break
				if rejection is not None:
					break
			if rejection is not None:
				continue
			# Check user-specified requirements
			for req in activeReqs:
				if not req(sample):
					rejection = 'user-specified requirement'
					break

		# obtained a valid sample; assemble a scene from it
		sampledObjects = tuple(sample[obj] for obj in objects)
		sampledParams = {}
		for param, value in self.params.items():
			sampledValue = sample[value] if isinstance(value, Samplable) else value
			assert not needsLazyEvaluation(sampledValue)
			sampledParams[param] = sampledValue
		scene = Scene(self.workspace, sampledObjects, ego, sampledParams)
		return scene, iterations

	def resetExternalSampler(self):
		"""Reset the scenario's external sampler, if any.

		If the Python random seed is reset before calling this function, this
		should cause the sequence of generated scenes to be deterministic."""
		self.externalSampler = ExternalSampler.forParameters(self.externalParams, self.params)



"""Support code for the CARLA world model."""

from scenic.simulators.formats.opendrive import OpenDriveWorkspace

class CarlaWorkspace(OpenDriveWorkspace):
    @property
    def minimumZoomSize(self):
        return 100



"""Workspaces based on OpenDRIVE maps."""

from .xodr_parser import RoadMap
from scenic.core.workspaces import Workspace
from scenic.core.regions import regionFromShapelyObject, nowhere
from scenic.core.vectors import VectorField

class OpenDriveWorkspace(Workspace):
    def __init__(self, path, n=20):
        '''Initialize from OpenDRIVE file at @path, with
        @n points per lane section reference line.'''
        self.road_map = RoadMap()
        self.road_map.parse(path)
        self.road_map.calculate_geometry(n, calc_intersect=True)
        drivable_poly = self.road_map.drivable_region
        sidewalk_poly = self.road_map.sidewalk_region
        intersect_poly = self.road_map.intersection_region
        self.road_direction = VectorField('Road Direction',
                                          self.road_map.heading_at)
        self.drivable_region = regionFromShapelyObject(drivable_poly,
                                                       orientation=self.road_direction)
        self.sidewalk_region = regionFromShapelyObject(sidewalk_poly)
        self.intersection_region = regionFromShapelyObject(intersect_poly,
                                                           orientation=self.road_direction)
        super().__init__()

        # lane_sec_dict is dict of road id to list of dict of lane id to Region.
        self.lane_sec_dict = {}
        for id_ in self.road_map.roads:
            lane_dicts = []
            for d in self.road_map.roads[id_].sec_lane_polys:
                lane_dicts.append({i: regionFromShapelyObject(d[i],
                                                              orientation=self.road_direction)
                                   for i in d.keys()})
            self.lane_sec_dict[id_] = lane_dicts

    def show(self, plt):
        self.drivable_region.show(plt)
        self.sidewalk_region.show(plt, style='b')
        self.intersection_region.show(plt, style='g')



"""Workspaces."""

from scenic.core.distributions import needsSampling
from scenic.core.regions import Region, everywhere
from scenic.core.geometry import findMinMax
from scenic.core.utils import RuntimeParseError

class Workspace(Region):
	"""A workspace describing the fixed world of a scenario"""
	def __init__(self, region=everywhere):
		if needsSampling(region):
			raise RuntimeParseError('workspace region must be fixed')
		super().__init__('workspace', orientation=region.orientation)
		self.region = region

	def show(self, plt):
		"""Render a schematic of the workspace for debugging"""
		try:
			aabb = self.region.getAABB()
		except NotImplementedError:		# unbounded Regions don't support this
			return
		((xmin, ymin), (xmax, ymax)) = aabb
		plt.xlim(xmin, xmax)
		plt.ylim(ymin, ymax)
		plt.gca().set_aspect('equal')

	def zoomAround(self, plt, objects, expansion=2):
		"""Zoom the schematic around the specified objects"""
		positions = (self.scenicToSchematicCoords(obj.position) for obj in objects)
		x, y = zip(*positions)
		minx, maxx = findMinMax(x)
		miny, maxy = findMinMax(y)
		sx = expansion * (maxx - minx)
		sy = expansion * (maxy - miny)
		s = max(sx, sy, self.minimumZoomSize) / 2.0
		s += max(max(obj.width, obj.height) for obj in objects)	# TODO improve
		cx = (maxx + minx) / 2.0
		cy = (maxy + miny) / 2.0
		plt.xlim(cx - s, cx + s)
		plt.ylim(cy - s, cy + s)

	@property
	def minimumZoomSize(self):
		return 0

	def scenicToSchematicCoords(self, coords):
		"""Convert Scenic coordinates to those used for schematic rendering."""
		return coords

	def uniformPointInner(self):
		return self.region.uniformPointInner()

	def intersect(self, other, triedReversed=False):
		return self.region.intersect(other, triedReversed)

	def containsPoint(self, point):
		return self.region.containsPoint(point)

	def containsObject(self, obj):
		return self.region.containsObject(obj)

	def getAABB(self):
		return self.region.getAABB()

	def __str__(self):
		return f'<Workspace on {self.region}>'

	def __eq__(self, other):
		if type(other) is not Workspace:
			return NotImplemented
		return other.region == self.region

	def __hash__(self):
		return hash(self.region)




"""Python implementations of Scenic language constructs.

This module is automatically imported by all Scenic programs. In addition to
defining the built-in functions, operators, specifiers, etc., it also stores
global state such as the list of all created Scenic objects.
"""

__all__ = (
	# Primitive statements and functions
	'ego', 'require', 'resample', 'param', 'mutate', 'verbosePrint',
	'sin', 'cos', 'hypot', 'max', 'min',
	# Prefix operators
	'Visible',
	'Front', 'Back', 'Left', 'Right',
	'FrontLeft', 'FrontRight', 'BackLeft', 'BackRight',
	# Infix operators
	'FieldAt', 'RelativeTo', 'OffsetAlong', 'RelativePosition',
	'RelativeHeading', 'ApparentHeading',
	'DistanceFrom', 'AngleTo', 'AngleFrom', 'Follow', 'CanSee',
	# Primitive types
	'Vector', 'VectorField', 'PolygonalVectorField',
	'Region', 'PointSetRegion', 'RectangularRegion', 'PolygonalRegion', 'PolylineRegion',
	'Workspace', 'Mutator',
	'Range', 'Options', 'Uniform', 'Discrete', 'Normal',
	'VerifaiParameter', 'VerifaiRange', 'VerifaiDiscreteRange', 'VerifaiOptions',
	# Constructible types
	'Point', 'OrientedPoint', 'Object',
	# Specifiers
	'With',
	'At', 'In', 'Beyond', 'VisibleFrom', 'VisibleSpec', 'OffsetBy', 'OffsetAlongSpec',
	'Facing', 'FacingToward', 'ApparentlyFacing',
	'LeftSpec', 'RightSpec', 'Ahead', 'Behind',
	'Following',
	# Constants
	'everywhere', 'nowhere',
	# Temporary stuff... # TODO remove
	'PropertyDefault'
)

# various Python types and functions used in the language but defined elsewhere
from scenic.core.geometry import sin, cos, hypot, max, min
from scenic.core.vectors import Vector, VectorField, PolygonalVectorField
from scenic.core.regions import (Region, PointSetRegion, RectangularRegion,
	PolygonalRegion, PolylineRegion, everywhere, nowhere)
from scenic.core.workspaces import Workspace
from scenic.core.distributions import Range, Options, Normal
Uniform = lambda *opts: Options(opts)		# TODO separate these?
Discrete = Options
from scenic.core.external_params import (VerifaiParameter, VerifaiRange, VerifaiDiscreteRange,
                                         VerifaiOptions)
from scenic.core.object_types import Mutator, Point, OrientedPoint, Object
from scenic.core.specifiers import PropertyDefault	# TODO remove

# everything that should not be directly accessible from the language is imported here:
import inspect
from scenic.core.distributions import Distribution, toDistribution
from scenic.core.type_support import isA, toType, toTypes, toScalar, toHeading, toVector
from scenic.core.type_support import evaluateRequiringEqualTypes, underlyingType
from scenic.core.geometry import RotatedRectangle, normalizeAngle, apparentHeadingAtPoint
from scenic.core.object_types import Constructible
from scenic.core.specifiers import Specifier
from scenic.core.lazy_eval import DelayedArgument
from scenic.core.utils import RuntimeParseError
from scenic.core.external_params import ExternalParameter

### Internals

activity = 0
evaluatingRequirement = False
allObjects = []		# ordered for reproducibility
egoObject = None
globalParameters = {}
externalParameters = []		# ordered for reproducibility
pendingRequirements = {}
inheritedReqs = []		# TODO improve handling of these?

def isActive():
	"""Are we in the middle of compiling a Scenic module?

	The 'activity' global can be >1 when Scenic modules in turn import other
	Scenic modules."""
	return activity > 0

def activate():
	"""Activate the veneer when beginning to compile a Scenic module."""
	global activity
	activity += 1
	assert not evaluatingRequirement

def deactivate():
	"""Deactivate the veneer after compiling a Scenic module."""
	global activity, allObjects, egoObject, globalParameters, externalParameters
	global pendingRequirements, inheritedReqs
	activity -= 1
	assert activity >= 0
	assert not evaluatingRequirement
	allObjects = []
	egoObject = None
	globalParameters = {}
	externalParameters = []
	pendingRequirements = {}
	inheritedReqs = []

def registerObject(obj):
	"""Add a Scenic object to the global list of created objects.

	This is called by the Object constructor."""
	if activity > 0:
		assert not evaluatingRequirement
		assert isinstance(obj, Constructible)
		allObjects.append(obj)
	elif evaluatingRequirement:
		raise RuntimeParseError('tried to create an object inside a requirement')

def registerExternalParameter(value):
	"""Register a parameter whose value is given by an external sampler."""
	if activity > 0:
		assert isinstance(value, ExternalParameter)
		externalParameters.append(value)

### Primitive statements and functions

def ego(obj=None):
	"""Function implementing loads and stores to the 'ego' pseudo-variable.

	The translator calls this with no arguments for loads, and with the source
	value for stores.
	"""
	global egoObject
	if obj is None:
		if egoObject is None:
			raise RuntimeParseError('referred to ego object not yet assigned')
	elif not isinstance(obj, Object):
		raise RuntimeParseError('tried to make non-object the ego object')
	else:
		egoObject = obj
	return egoObject

def require(reqID, req, line, prob=1):
	"""Function implementing the require statement."""
	if evaluatingRequirement:
		raise RuntimeParseError('tried to create a requirement inside a requirement')
	# the translator wrapped the requirement in a lambda to prevent evaluation,
	# so we need to save the current values of all referenced names; throw in
	# the ego object too since it can be referred to implicitly
	assert reqID not in pendingRequirements
	pendingRequirements[reqID] = (req, getAllGlobals(req), egoObject, line, prob)

def getAllGlobals(req, restrictTo=None):
	"""Find all names the given lambda depends on, along with their current bindings."""
	namespace = req.__globals__
	if restrictTo is not None and restrictTo is not namespace:
		return {}
	externals = inspect.getclosurevars(req)
	assert not externals.nonlocals		# TODO handle these
	globs = dict(externals.builtins)
	for name, value in externals.globals.items():
		globs[name] = value
		if inspect.isfunction(value):
			subglobs = getAllGlobals(value, restrictTo=namespace)
			for name, value in subglobs.items():
				if name in globs:
					assert value is globs[name]
				else:
					globs[name] = value
	return globs

def resample(dist):
	"""The built-in resample function."""
	return dist.clone() if isinstance(dist, Distribution) else dist

def verbosePrint(msg):
	"""Built-in function printing a message when the verbosity is >0."""
	import scenic.syntax.translator as translator
	if translator.verbosity >= 1:
		indent = '  ' * activity if translator.verbosity >= 2 else '  '
		print(indent + msg)

def param(*quotedParams, **params):
	"""Function implementing the param statement."""
	if evaluatingRequirement:
		raise RuntimeParseError('tried to create a global parameter inside a requirement')
	for name, value in params.items():
		globalParameters[name] = toDistribution(value)
	assert len(quotedParams) % 2 == 0, quotedParams
	it = iter(quotedParams)
	for name, value in zip(it, it):
		globalParameters[name] = toDistribution(value)

def mutate(*objects):		# TODO update syntax
	"""Function implementing the mutate statement."""
	if evaluatingRequirement:
		raise RuntimeParseError('used mutate statement inside a requirement')
	if len(objects) == 0:
		objects = allObjects
	for obj in objects:
		if not isinstance(obj, Object):
			raise RuntimeParseError('"mutate X" with X not an object')
		obj.mutationEnabled = True

### Prefix operators

def Visible(region):
	"""The 'visible <region>' operator."""
	if not isinstance(region, Region):
		raise RuntimeParseError('"visible X" with X not a Region')
	return region.intersect(ego().visibleRegion)

# front of <object>, etc.
ops = (
	'front', 'back', 'left', 'right',
	'front left', 'front right',
	'back left', 'back right'
)
template = '''\
def {function}(X):
	"""The '{syntax} of <object>' operator."""
	if not isinstance(X, Object):
		raise RuntimeParseError('"{syntax} of X" with X not an Object')
	return X.{property}
'''
for op in ops:
	func = ''.join(word.capitalize() for word in op.split(' '))
	prop = func[0].lower() + func[1:]
	definition = template.format(function=func, syntax=op, property=prop)
	exec(definition)

### Infix operators

def FieldAt(X, Y):
	"""The '<VectorField> at <vector>' operator."""
	if not isinstance(X, VectorField):
		raise RuntimeParseError('"X at Y" with X not a vector field')
	Y = toVector(Y, '"X at Y" with Y not a vector')
	return X[Y]

def RelativeTo(X, Y):
	"""The 'X relative to Y' polymorphic operator.

	Allowed forms:
		F relative to G (with at least one a field, the other a field or heading)
		<vector> relative to <oriented point> (and vice versa)
		<vector> relative to <vector>
		<heading> relative to <heading>
	"""
	xf, yf = isA(X, VectorField), isA(Y, VectorField)
	if xf or yf:
		if xf and yf and X.valueType != Y.valueType:
			raise RuntimeParseError('"X relative to Y" with X, Y fields of different types')
		fieldType = X.valueType if xf else Y.valueType
		error = '"X relative to Y" with field and value of different types'
		def helper(context):
			pos = context.position.toVector()
			xp = X[pos] if xf else toType(X, fieldType, error)
			yp = Y[pos] if yf else toType(Y, fieldType, error)
			return xp + yp
		return DelayedArgument({'position'}, helper)
	else:
		if isinstance(X, OrientedPoint):	# TODO too strict?
			if isinstance(Y, OrientedPoint):
				raise RuntimeParseError('"X relative to Y" with X, Y both oriented points')
			Y = toVector(Y, '"X relative to Y" with X an oriented point but Y not a vector')
			return X.relativize(Y)
		elif isinstance(Y, OrientedPoint):
			X = toVector(X, '"X relative to Y" with Y an oriented point but X not a vector')
			return Y.relativize(X)
		else:
			X = toTypes(X, (Vector, float), '"X relative to Y" with X neither a vector nor scalar')
			Y = toTypes(Y, (Vector, float), '"X relative to Y" with Y neither a vector nor scalar')
			return evaluateRequiringEqualTypes(lambda: X + Y, X, Y,
			                                   '"X relative to Y" with vector and scalar')

def OffsetAlong(X, H, Y):
	"""The 'X offset along H by Y' polymorphic operator.

	Allowed forms:
		<vector> offset along <heading> by <vector>
		<vector> offset along <field> by <vector>
	"""
	X = toVector(X, '"X offset along H by Y" with X not a vector')
	Y = toVector(Y, '"X offset along H by Y" with Y not a vector')
	if isinstance(H, VectorField):
		H = H[X]
	H = toHeading(H, '"X offset along H by Y" with H not a heading or vector field')
	return X.offsetRotated(H, Y)

def RelativePosition(X, Y=None):
	"""The 'relative position of <vector> [from <vector>]' operator.

	If the 'from <vector>' is omitted, the position of ego is used.
	"""
	X = toVector(X, '"relative position of X from Y" with X not a vector')
	if Y is None:
		Y = ego()
	Y = toVector(Y, '"relative position of X from Y" with Y not a vector')
	return X - Y

def RelativeHeading(X, Y=None):
	"""The 'relative heading of <heading> [from <heading>]' operator.

	If the 'from <heading>' is omitted, the heading of ego is used.
	"""
	X = toHeading(X, '"relative heading of X from Y" with X not a heading')
	if Y is None:
		Y = ego().heading
	else:
		Y = toHeading(Y, '"relative heading of X from Y" with Y not a heading')
	return normalizeAngle(X - Y)

def ApparentHeading(X, Y=None):
	"""The 'apparent heading of <oriented point> [from <vector>]' operator.

	If the 'from <vector>' is omitted, the position of ego is used.
	"""
	if not isinstance(X, OrientedPoint):
		raise RuntimeParseError('"apparent heading of X from Y" with X not an OrientedPoint')
	if Y is None:
		Y = ego()
	Y = toVector(Y, '"relative heading of X from Y" with Y not a vector')
	return apparentHeadingAtPoint(X.position, X.heading, Y)

def DistanceFrom(X, Y=None):
	"""The 'distance from <vector> [to <vector>]' operator.

	If the 'to <vector>' is omitted, the position of ego is used.
	"""
	X = toVector(X, '"distance from X to Y" with X not a vector')
	if Y is None:
		Y = ego()
	Y = toVector(Y, '"distance from X to Y" with Y not a vector')
	return X.distanceTo(Y)

def AngleTo(X):
	"""The 'angle to <vector>' operator (using the position of ego as the reference)."""
	X = toVector(X, '"angle to X" with X not a vector')
	return ego().angleTo(X)

def AngleFrom(X, Y):
	"""The 'angle from <vector> to <vector>' operator."""
	X = toVector(X, '"angle from X to Y" with X not a vector')
	Y = toVector(Y, '"angle from X to Y" with Y not a vector')
	return X.angleTo(Y)

def Follow(F, X, D):
	"""The 'follow <field> from <vector> for <number>' operator."""
	if not isinstance(F, VectorField):
		raise RuntimeParseError('"follow F from X for D" with F not a vector field')
	X = toVector(X, '"follow F from X for D" with X not a vector')
	D = toScalar(D, '"follow F from X for D" with D not a number')
	pos = F.followFrom(X, D)
	heading = F[pos]
	return OrientedPoint(position=pos, heading=heading)

def CanSee(X, Y):
	"""The 'X can see Y' polymorphic operator.

	Allowed forms:
		<point> can see <object>
		<point> can see <vector>
	"""
	if not isinstance(X, Point):
		raise RuntimeParseError('"X can see Y" with X not a Point')
	if isinstance(Y, Point):
		return X.canSee(Y)
	else:
		Y = toVector(Y, '"X can see Y" with Y not a vector')
		return X.visibleRegion.containsPoint(Y)

### Specifiers

def With(prop, val):
	"""The 'with <property> <value>' specifier.

	Specifies the given property, with no dependencies.
	"""
	return Specifier(prop, val)

def At(pos):
	"""The 'at <vector>' specifier.

	Specifies 'position', with no dependencies."""
	pos = toVector(pos, 'specifier "at X" with X not a vector')
	return Specifier('position', pos)

def In(region):
	"""The 'in/on <region>' specifier.

	Specifies 'position', with no dependencies. Optionally specifies 'heading'
	if the given Region has a preferred orientation.
	"""
	region = toType(region, Region, 'specifier "in/on R" with R not a Region')
	extras = {'heading'} if alwaysProvidesOrientation(region) else {}
	return Specifier('position', Region.uniformPointIn(region), optionals=extras)

def alwaysProvidesOrientation(region):
	"""Whether a Region or distribution over Regions always provides an orientation."""
	if isinstance(region, Region):
		return region.orientation is not None
	elif isinstance(region, Options):
		return all(alwaysProvidesOrientation(opt) for opt in region.options)
	else:
		return False

def Beyond(pos, offset, fromPt=None):
	"""The 'beyond X by Y [from Z]' polymorphic specifier.

	Specifies 'position', with no dependencies.

	Allowed forms:
		beyond <vector> by <number> [from <vector>]
		beyond <vector> by <vector> [from <vector>]

	If the 'from <vector>' is omitted, the position of ego is used.
	"""
	pos = toVector(pos, 'specifier "beyond X by Y" with X not a vector')
	dType = underlyingType(offset)
	if dType is float or dType is int:
		offset = Vector(0, offset)
	elif dType is not Vector:
		raise RuntimeParseError('specifier "beyond X by Y" with Y not a number or vector')
	if fromPt is None:
		fromPt = ego()
	fromPt = toVector(fromPt, 'specifier "beyond X by Y from Z" with Z not a vector')
	lineOfSight = fromPt.angleTo(pos)
	return Specifier('position', pos.offsetRotated(lineOfSight, offset))

def VisibleFrom(base):
	"""The 'visible from <Point>' specifier.

	Specifies 'position', with no dependencies.

	This uses the given object's 'visibleRegion' property, and so correctly
	handles the view regions of Points, OrientedPoints, and Objects.
	"""
	if not isinstance(base, Point):
		raise RuntimeParseError('specifier "visible from O" with O not a Point')
	return Specifier('position', Region.uniformPointIn(base.visibleRegion))

def VisibleSpec():
	"""The 'visible' specifier (equivalent to 'visible from ego').

	Specifies 'position', with no dependencies.
	"""
	return VisibleFrom(ego())

def OffsetBy(offset):
	"""The 'offset by <vector>' specifier.

	Specifies 'position', with no dependencies.
	"""
	offset = toVector(offset, 'specifier "offset by X" with X not a vector')
	pos = RelativeTo(offset, ego()).toVector()
	return Specifier('position', pos)

def OffsetAlongSpec(direction, offset):
	"""The 'offset along X by Y' polymorphic specifier.

	Specifies 'position', with no dependencies.

	Allowed forms:
		offset along <heading> by <vector>
		offset along <field> by <vector>
	"""
	return Specifier('position', OffsetAlong(ego(), direction, offset))

def Facing(heading):
	"""The 'facing X' polymorphic specifier.

	Specifies 'heading', with dependencies depending on the form:
		facing <number> -- no dependencies;
		facing <field> -- depends on 'position'.
	"""
	if isinstance(heading, VectorField):
		return Specifier('heading', DelayedArgument({'position'},
		                                            lambda self: heading[self.position]))
	else:
		heading = toHeading(heading, 'specifier "facing X" with X not a heading or vector field')
		return Specifier('heading', heading)

def FacingToward(pos):
	"""The 'facing toward <vector>' specifier.

	Specifies 'heading', depending on 'position'.
	"""
	pos = toVector(pos, 'specifier "facing toward X" with X not a vector')
	return Specifier('heading', DelayedArgument({'position'},
	                                            lambda self: self.position.angleTo(pos)))

def ApparentlyFacing(heading, fromPt=None):
	"""The 'apparently facing <heading> [from <vector>]' specifier.

	Specifies 'heading', depending on 'position'.

	If the 'from <vector>' is omitted, the position of ego is used.
	"""
	heading = toHeading(heading, 'specifier "apparently facing X" with X not a heading')
	if fromPt is None:
		fromPt = ego()
	fromPt = toVector(fromPt, 'specifier "apparently facing X from Y" with Y not a vector')
	value = lambda self: fromPt.angleTo(self.position) + heading
	return Specifier('heading', DelayedArgument({'position'}, value))

def LeftSpec(pos, dist=0):
	"""The 'left of X [by Y]' polymorphic specifier.

	Specifies 'position', depending on 'width'. See other dependencies below.

	Allowed forms:
		left of <oriented point> [by <scalar/vector>] -- optionally specifies 'heading';
		left of <vector> [by <scalar/vector>] -- depends on 'heading'.

	If the 'by <scalar/vector>' is omitted, zero is used.
	"""
	return leftSpecHelper('left of', pos, dist, 'width', lambda dist: (dist, 0),
	                      lambda self, dx, dy: Vector(-self.width / 2 - dx, dy))

def RightSpec(pos, dist=0):
	"""The 'right of X [by Y]' polymorphic specifier.

	Specifies 'position', depending on 'width'. See other dependencies below.

	Allowed forms:
		right of <oriented point> [by <scalar/vector>] -- optionally specifies 'heading';
		right of <vector> [by <scalar/vector>] -- depends on 'heading'.

	If the 'by <scalar/vector>' is omitted, zero is used.
	"""
	return leftSpecHelper('right of', pos, dist, 'width', lambda dist: (dist, 0),
	                      lambda self, dx, dy: Vector(self.width / 2 + dx, dy))

def Ahead(pos, dist=0):
	"""The 'ahead of X [by Y]' polymorphic specifier.

	Specifies 'position', depending on 'height'. See other dependencies below.

	Allowed forms:

	* ``ahead of`` <oriented point> [``by`` <scalar/vector>] -- optionally specifies 'heading';
	* ``ahead of`` <vector> [``by`` <scalar/vector>] -- depends on 'heading'.

	If the 'by <scalar/vector>' is omitted, zero is used.
	"""
	return leftSpecHelper('ahead of', pos, dist, 'height', lambda dist: (0, dist),
	                      lambda self, dx, dy: Vector(dx, self.height / 2 + dy))

def Behind(pos, dist=0):
	"""The 'behind X [by Y]' polymorphic specifier.

	Specifies 'position', depending on 'height'. See other dependencies below.

	Allowed forms:
		behind <oriented point> [by <scalar/vector>] -- optionally specifies 'heading';
		behind <vector> [by <scalar/vector>] -- depends on 'heading'.

	If the 'by <scalar/vector>' is omitted, zero is used.
	"""
	return leftSpecHelper('behind', pos, dist, 'height', lambda dist: (0, dist),
	                      lambda self, dx, dy: Vector(dx, -self.height / 2 - dy))

def leftSpecHelper(syntax, pos, dist, axis, toComponents, makeOffset):
	extras = set()
	dType = underlyingType(dist)
	if dType is float or dType is int:
		dx, dy = toComponents(dist)
	elif dType is Vector:
		dx, dy = dist
	else:
		raise RuntimeParseError(f'"{syntax} X by D" with D not a number or vector')
	if isinstance(pos, OrientedPoint):		# TODO too strict?
		val = lambda self: pos.relativePosition(makeOffset(self, dx, dy))
		new = DelayedArgument({axis}, val)
		extras.add('heading')
	else:
		pos = toVector(pos, f'specifier "{syntax} X" with X not a vector')
		val = lambda self: pos.offsetRotated(self.heading, makeOffset(self, dx, dy))
		new = DelayedArgument({axis, 'heading'}, val)
	return Specifier('position', new, optionals=extras)

def Following(field, dist, fromPt=None):
	"""The 'following F [from X] for D' specifier.

	Specifies 'position', and optionally 'heading', with no dependencies.

	Allowed forms:
		following <field> [from <vector>] for <number>

	If the 'from <vector>' is omitted, the position of ego is used.
	"""
	if fromPt is None:
		fromPt = ego()
	else:
		dist, fromPt = fromPt, dist
	if not isinstance(field, VectorField):
		raise RuntimeParseError('"following F" specifier with F not a vector field')
	fromPt = toVector(fromPt, '"following F from X for D" with X not a vector')
	dist = toScalar(dist, '"following F for D" with D not a number')
	pos = field.followFrom(fromPt, dist)
	heading = field[pos]
	val = OrientedPoint(position=pos, heading=heading)
	return Specifier('position', val, optionals={'heading'})



"""Python library supporting the main Scenic module."""

import time
import sys
import math
import itertools

import numpy as np
import shapely.geometry
import shapely.ops

import scenic.simulators.webots.world_parser as world_parser
from scenic.simulators.webots.common import webotsToScenicPosition, webotsToScenicRotation
from scenic.core.workspaces import Workspace
from scenic.core.vectors import PolygonalVectorField
from scenic.core.regions import PolygonalRegion, PolylineRegion, nowhere
from scenic.core.geometry import (normalizeAngle, rotateVector, polygonUnion, cleanChain,
                                  plotPolygon)
from scenic.syntax.veneer import verbosePrint

def polygonWithPoints(points):
	polygon = shapely.geometry.Polygon(points)
	if not polygon.is_valid:	# TODO improve hack?
		verbosePrint(f'WARNING: simplifying invalid polygon with points {points}')
		polygon = polygon.simplify(0.5)
		if not polygon.is_valid:
			raise RuntimeError(f'unable to simplify polygon {polygon}')
	return polygon

def regionWithPolygons(polygons, orientation=None):
	if len(polygons) == 0:
		return nowhere
	else:
		return PolygonalRegion(polygon=polygons, orientation=orientation)

## Classes for WBT nodes we are interested in

class OSMObject:
	"""Objects with OSM id tags"""
	def __init__(self, attrs):
		self.attrs = attrs
		self.osmID = attrs['id']

class Road(OSMObject):
	"""OSM roads"""
	def __init__(self, attrs, driveOnLeft=False):
		super().__init__(attrs)
		self.driveOnLeft = driveOnLeft
		self.translation = attrs['translation']
		pts = [np.array(webotsToScenicPosition(p + self.translation)) for p in attrs['wayPoints']]
		self.waypoints = tuple(cleanChain(pts, 0.05))
		assert len(self.waypoints) > 1, pts
		self.width = float(attrs.get('width', 7))
		self.lanes = int(attrs.get('numberOfLanes', 2))
		if self.lanes < 1:
			raise RuntimeError(f'Road {self.osmID} has fewer than 1 lane!')
		self.forwardLanes = int(attrs.get('numberOfForwardLanes', 1))
		# if self.forwardLanes < 1:
		# 	raise RuntimeError(f'Road {self.osmID} has fewer than 1 forward lane!')
		self.backwardLanes = self.lanes - self.forwardLanes
		self.hasLeftSidewalk = attrs.get('leftBorder', True)
		self.hasRightSidewalk = attrs.get('rightBorder', True)
		self.sidewalkWidths = list(attrs.get('roadBorderWidth', [0.8]))
		if ((self.hasLeftSidewalk or self.hasRightSidewalk)
				and len(self.sidewalkWidths) < 1):
			raise RuntimeError(f'Road {self.osmID} has sidewalk with empty width!')
		self.startCrossroad = attrs.get('startJunction')
		self.endCrossroad = attrs.get('endJunction')

	def computeGeometry(self, crossroads, snapTolerance=0.05):
		## Approximate bounding polygon and sidewalks
		# TODO improve this!!!
		lefts, rights = [], []
		leftSidewalk, rightSidewalk = [], []
		headings = []
		sidewalkWidths = itertools.chain(self.sidewalkWidths,
		                                 itertools.repeat(self.sidewalkWidths[-1]))
		segments = zip(self.waypoints, sidewalkWidths)
		for i, segment in enumerate(segments):
			point, sidewalkWidth = segment
			if i+1 < len(self.waypoints):
				nextPt = self.waypoints[i+1]
				dx, dy = nextPt - point
				heading = normalizeAngle(math.atan2(dy, dx) - (math.pi / 2))
				headings.append(heading)
				perp = np.array([-dy, dx])
				perp /= np.linalg.norm(perp)
			else:
				pass	# use perp from last segment
			toEdge = perp * (self.width / 2)
			left = point + toEdge
			right = point - toEdge
			lefts.append(left)
			rights.append(right)
			toEdge = perp * sidewalkWidth
			leftSidewalk.append(left + toEdge)
			rightSidewalk.append(right - toEdge)

		# Snap to adjacent crossroads if possible
		if snapTolerance > 0:
			sc = self.startCrossroad
			if sc is not None:
				if sc not in crossroads:
					raise RuntimeError(f'Road {self.osmID} begins at invalid crossroad {sc}')
				crossroad = crossroads[sc]
				if crossroad.region is not None:
					pt = shapely.geometry.Point(lefts[0])
					pt = shapely.ops.snap(pt, crossroad.region.polygons, snapTolerance)
					lefts[0] = np.array([pt.x, pt.y])
					pt = shapely.geometry.Point(rights[0])
					pt = shapely.ops.snap(pt, crossroad.region.polygons, snapTolerance)
					rights[0] = np.array([pt.x, pt.y])
					perp = lefts[0] - rights[0]
					toEdge = perp * (self.sidewalkWidths[0] / np.linalg.norm(perp))
					leftSidewalk[0] = lefts[0] + toEdge
					rightSidewalk[0] = rights[0] - toEdge
			ec = self.endCrossroad
			if ec is not None:
				if ec not in crossroads:
					raise RuntimeError(f'Road {self.osmID} ends at invalid crossroad {ec}')
				crossroad = crossroads[ec]
				if crossroad.region is not None:
					pt = shapely.geometry.Point(lefts[-1])
					pt = shapely.ops.snap(pt, crossroad.region.polygons, snapTolerance)
					lefts[-1] = np.array([pt.x, pt.y])
					pt = shapely.geometry.Point(rights[-1])
					pt = shapely.ops.snap(pt, crossroad.region.polygons, snapTolerance)
					rights[-1] = np.array([pt.x, pt.y])
					perp = lefts[-1] - rights[-1]
					toEdge = perp * (self.sidewalkWidths[-1] / np.linalg.norm(perp))
					leftSidewalk[-1] = lefts[-1] + toEdge
					rightSidewalk[-1] = rights[-1] - toEdge

		roadPoints = lefts + list(reversed(rights))
		self.leftCurb = PolylineRegion(reversed(lefts))
		self.rightCurb = PolylineRegion(rights)

		self.leftSidewalk = self.rightSidewalk = None
		if self.hasLeftSidewalk:
			points = lefts + list(reversed(leftSidewalk))
			polygon = polygonWithPoints(points)
			assert polygon.is_valid, self.waypoints
			self.leftSidewalk = PolygonalRegion(polygon=polygon)
		if self.hasRightSidewalk:
			points = rights + list(reversed(rightSidewalk))
			polygon = polygonWithPoints(points)
			assert polygon.is_valid, self.waypoints
			self.rightSidewalk = PolygonalRegion(polygon=polygon)

		## Compute lanes and traffic directions
		cells = []
		la, ra = lefts[0], rights[0]
		gapA = (ra - la) / self.lanes
		markerA = ra
		laneMarkers = [[] for lane in range(self.lanes)]
		for lb, rb, heading in zip(lefts[1:], rights[1:], headings):
			# Compute lanes for this segment of road
			gapB = (rb - lb) / self.lanes
			markerB = rb
			for lane, markers in enumerate(laneMarkers):
				forward = lane < self.forwardLanes
				if self.driveOnLeft:
					forward = not forward
				nextMarkerA = markerA - gapA
				nextMarkerB = markerB - gapB
				markers.append(nextMarkerA)
				cell = shapely.geometry.Polygon((markerA, markerB, nextMarkerB, nextMarkerA))
				heading = heading if forward else normalizeAngle(heading + math.pi)
				cells.append((cell, heading))
				markerA = nextMarkerA
				markerB = nextMarkerB
			gapA = gapB
			markerA = rb
		self.lanes = []
		markerB = rb
		rightEdge = rights
		for lane, markers in enumerate(laneMarkers):
			markerB = markerB - gapB
			markers.append(markerB)
			self.lanes.append(PolygonalRegion(rightEdge + list(reversed(markers))))
			rightEdge = markers
		self.laneMarkers = laneMarkers[:-1]
		self.cells = cells
		self.direction = PolygonalVectorField(f'Road{self.osmID}Direction', cells)

		roadPolygon = polygonWithPoints(roadPoints)
		self.region = PolygonalRegion(polygon=roadPolygon, orientation=self.direction)

	def show(self, plt):
		if self.hasLeftSidewalk:
			x, y = zip(*self.leftSidewalk.points)
			plt.fill(x, y, '#A0A0FF')
		if self.hasRightSidewalk:
			x, y = zip(*self.rightSidewalk.points)
			plt.fill(x, y, '#A0A0FF')
		self.region.show(plt, style='r:')
		x, y = zip(*self.lanes[0].points)
		plt.fill(x, y, color=(0.8, 1.0, 0.8))
		for lane, markers in enumerate(self.laneMarkers):
			x, y = zip(*markers)
			color = (0.8, 0.8, 0) if lane == self.backwardLanes - 1 else (0.3, 0.3, 0.3)
			plt.plot(x, y, '--', color=color)

class Crossroad(OSMObject):
	"""OSM crossroads"""
	def __init__(self, attrs):
		super().__init__(attrs)
		self.translation = attrs['translation']
		points = list(np.array(webotsToScenicPosition(p + self.translation))
		              for p in attrs['shape'])
		if len(points) > 0:
			self.points = points
			self.region = PolygonalRegion(self.points)
		else:
			verbosePrint(f'WARNING: Crossroad {self.osmID} has empty shape field!')
			self.region = None

	def show(self, plt):
		if self.region is not None:
			x, y = zip(*self.points)
			plt.fill(x, y, color=(1, 0.9, 0.9))
			plt.plot(x, y, ':', color=(1, 0.5, 0))

class PedestrianCrossing:
	"""PedestrianCrossing nodes"""
	def __init__(self, attrs):
		self.translation = attrs.get('translation', np.array((0, 0, 0)))
		pos = np.array(webotsToScenicPosition(self.translation))
		name = attrs.get('name', '')
		self.angle = webotsToScenicRotation(attrs.get('rotation', (0, 1, 0, 0)))
		if self.angle is None:
			raise RuntimeError(f'PedestrianCrossing "{name}" is not 2D!')
		size = attrs.get('size', (20, 8))
		self.length, self.width = float(size[0]), float(size[1])
		self.length += 0.2		# pad length to intersect sidewalks better	# TODO improve?
		hl, hw = self.length / 2, self.width / 2
		self.corners = tuple(pos + rotateVector(vec, -self.angle)
			for vec in ((hl, hw), (-hl, hw), (-hl, -hw), (hl, -hw)))
		self.region = PolygonalRegion(self.corners)

	def show(self, plt):
		x, y = zip(*self.corners)
		plt.fill(x, y, color='#A0A0FF')

## Workspace derived from a WBT world

class WebotsWorkspace(Workspace):
	def __init__(self, world):
		# Find roads, crossroads, and pedestrian crossings
		nodeClasses = {
			'Road': Road,
			'Crossroad': Crossroad,
			'PedestrianCrossing': PedestrianCrossing
		}
		self.roads, self.crossroads, self.crossings = world_parser.findNodeTypesIn(
			('Road', 'Crossroad', 'PedestrianCrossing'), world, nodeClasses)

		# Compute road geometry
		crossroadsByID = { crossroad.osmID: crossroad for crossroad in self.crossroads }
		for road in self.roads:
			road.computeGeometry(crossroadsByID)

		# Construct regions
		allCells = []
		drivableAreas = []
		for road in self.roads:
			assert road.region.polygons.is_valid, (road.waypoints, road.region.points)
			allCells.extend(road.cells)
		for crossroad in self.crossroads:
			if crossroad.region is not None:
				for poly in crossroad.region.polygons:
					allCells.append((poly, None))
		if not allCells:
			raise RuntimeError('Webots world has no drivable geometry!')
		self.roadDirection = PolygonalVectorField(
		    'roadDirection', allCells,
		    headingFunction=lambda pos: 0, defaultHeading=0		# TODO fix
		)
		if not self.roads:
			roadPoly = None
			self.roadsRegion = nowhere
			self.curbsRegion = nowhere
		else:
			roadPoly = polygonUnion(road.region.polygons for road in self.roads)
			self.roadsRegion = PolygonalRegion(polygon=roadPoly,
			                                   orientation=self.roadDirection)
			drivableAreas.append(roadPoly)
			curbs = [road.leftCurb.lineString for road in self.roads]
			curbs.extend(road.rightCurb.lineString for road in self.roads)
			curbLine = shapely.ops.unary_union(curbs)
			self.curbsRegion = PolylineRegion(polyline=curbLine)
		if not self.crossroads:
			crossroadPoly = None
			self.crossroadsRegion = nowhere
		else:
			crossroadPoly = polygonUnion(cr.region.polygons
			                             for cr in self.crossroads
			                             if cr.region is not None)
			self.crossroadsRegion = PolygonalRegion(polygon=crossroadPoly,
			                                        orientation=self.roadDirection)
			drivableAreas.append(crossroadPoly)

		sidewalks = []
		walkableAreas = []
		for road in self.roads:
			if road.hasLeftSidewalk:
				sidewalks.append(road.leftSidewalk.polygons)
			if road.hasRightSidewalk:
				sidewalks.append(road.rightSidewalk.polygons)
		if not sidewalks:
			sidewalksPoly = None
			self.sidewalksRegion = nowhere
		else:
			sidewalksPoly = polygonUnion(sidewalks)
			self.sidewalksRegion = regionWithPolygons(sidewalksPoly)
			walkableAreas.append(sidewalksPoly)

		if not self.crossings:
			crossingsPoly = None
			self.crossingsRegion = nowhere
		else:
			crossingsPoly = polygonUnion(crossing.region.polygons
			                             for crossing in self.crossings)
			self.crossingsRegion = regionWithPolygons(crossingsPoly)
			walkableAreas.append(crossingsPoly)

		if not walkableAreas:
			self.walkableRegion = nowhere
		else:
			walkablePoly = polygonUnion(walkableAreas)
			self.walkableRegion = regionWithPolygons(walkablePoly)
		drivablePoly = polygonUnion(drivableAreas)
		self.drivableRegion = PolygonalRegion(polygon=drivablePoly,
		                                      orientation=self.roadDirection)
		workspacePoly = polygonUnion(drivableAreas + walkableAreas)
		self.workspaceRegion = PolygonalRegion(polygon=workspacePoly)

		# Identify various roads and lanes of interest
		slowCurbs = []
		for road in self.roads:
			if road.forwardLanes > 1:
				slowCurbs.append(road.rightCurb)
			if road.backwardLanes > 1:
				slowCurbs.append(road.leftCurb)
		self.slowCurbs = slowCurbs

		super().__init__(self.workspaceRegion)

	def show(self, plt):
		plt.gca().set_aspect('equal')
		self.drivableRegion.show(plt)
		for road in self.roads:
			road.show(plt)
		for crossroad in self.crossroads:
			crossroad.show(plt)
		for crossing in self.crossings:
			crossing.show(plt)
		for curb in self.slowCurbs:
			curb.show(plt, style='b-')

	@property
	def minimumZoomSize(self):
		return 30
